# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import
from __future__ import division

import re

from dataproperty import DataProperty
import six
import typepy
from typepy.type import RealNumber

import subprocrunner as spr

from ._common import (
    logging_context,
    get_no_limit_kbits,
    sanitize_network,
    verify_network_interface,
    run_command_helper,
)
from ._const import (
    KILO_SIZE,
    LIST_MANGLE_TABLE_COMMAND,
    ShapingAlgorithm,
    Tc,
    TcCommandOutput,
    TrafficDirection,
)
from ._converter import (
    Humanreadable,
    HumanReadableTime,
)
from ._error import (
    NetworkInterfaceNotFoundError,
    InvalidParameterError,
    UnitNotFoundError,
)
from ._iptables import IptablesMangleController
from ._logger import logger
from .shaper.htb import HtbShaper
from .shaper.tbf import TbfShaper


def _validate_within_min_max(param_name, value, min_value, max_value, unit):
    if value is None:
        return

    if unit is None:
        unit = ""
    else:
        unit = "[{:s}]".format(unit)

    if value > max_value:
        raise InvalidParameterError(
            "'{:s}' is too high".format(param_name),
            expected="<={:s}{:s}".format(
                DataProperty(max_value).to_str(), unit),
            value="{:s}{:s}".format(
                DataProperty(value).to_str(), unit))

    if value < min_value:
        raise InvalidParameterError(
            "'{:s}' is too low".format(param_name),
            expected=">={:s}{:s}".format(
                DataProperty(min_value).to_str(), unit),
            value="{:s}{:s}".format(DataProperty(value).to_str(), unit))


class TrafficControl(object):
    MIN_PACKET_LOSS_RATE = 0  # [%]
    MAX_PACKET_LOSS_RATE = 100  # [%]

    MIN_PACKET_DUPLICATE_RATE = 0  # [%]
    MAX_PACKET_DUPLICATE_RATE = 100  # [%]

    MIN_CORRUPTION_RATE = 0  # [%]
    MAX_CORRUPTION_RATE = 100  # [%]

    MIN_REORDERING_RATE = 0  # [%]
    MAX_REORDERING_RATE = 100  # [%]

    __MIN_PORT = 0
    __MAX_PORT = 65535

    REGEXP_FILE_EXISTS = re.compile("RTNETLINK answers: File exists")

    EXISTS_MSG_TEMPLATE = (
        "{:s} try to execute with: "
        "(a) --overwrite option if you want to overwrite "
        "the existing rule. "
        "(b) --add option if you want to add a new rule in addition "
        "to the existing rules. "
        "(c) --change option if you want to change the existing "
        "rule parameter."
    )

    @property
    def device(self):
        return self.__device

    @property
    def ifb_device(self):
        return "ifb{:d}".format(self.__qdisc_major_id)

    @property
    def direction(self):
        return self.__direction

    @property
    def bandwidth_rate(self):
        # convert bandwidth string [K/M/G bit per second] to a number
        try:
            return Humanreadable(
                self.__bandwidth_rate, kilo_size=KILO_SIZE).to_kilo_bit()
        except (InvalidParameterError, UnitNotFoundError, TypeError):
            return None

    @property
    def latency_time(self):
        return self.__latency_time

    @property
    def latency_distro_time(self):
        return self.__latency_distro_time

    @property
    def packet_loss_rate(self):
        return self.__packet_loss_rate

    @property
    def packet_duplicate_rate(self):
        return self.__packet_duplicate_rate

    @property
    def corruption_rate(self):
        return self.__corruption_rate

    @property
    def reordering_rate(self):
        return self.__reordering_rate

    @property
    def dst_network(self):
        return self.__dst_network

    @property
    def exclude_dst_network(self):
        return self.__exclude_dst_network

    @property
    def src_network(self):
        return self.__src_network

    @property
    def exclude_src_network(self):
        return self.__exclude_src_network

    @property
    def src_port(self):
        return self.__src_port

    @property
    def exclude_src_port(self):
        return self.__exclude_src_port

    @property
    def dst_port(self):
        return self.__dst_port

    @property
    def exclude_dst_port(self):
        return self.__exclude_dst_port

    @property
    def is_change_shaping_rule(self):
        return self.__is_change_shaping_rule

    @property
    def is_add_shaping_rule(self):
        return self.__is_add_shaping_rule

    @property
    def is_enable_iptables(self):
        return self.__is_enable_iptables

    @property
    def qdisc_major_id(self):
        return self.__qdisc_major_id

    @property
    def qdisc_major_id_str(self):
        return "{:x}".format(self.__qdisc_major_id)

    @property
    def ip_version(self):
        return 6 if self.__is_ipv6 else 4

    @property
    def protocol(self):
        return "ipv6" if self.__is_ipv6 else "ip"

    @property
    def protocol_match(self):
        return "ip6" if self.__is_ipv6 else "ip"

    @property
    def tc_command_output(self):
        return self.__tc_command_output

    @property
    def iptables_ctrl(self):
        return self.__iptables_ctrl

    def __init__(
            self, device,
            direction=None, bandwidth_rate=None,
            latency_time=None, latency_distro_time=None,
            packet_loss_rate=None, packet_duplicate_rate=None,
            corruption_rate=None, reordering_rate=None,
            dst_network=None, exclude_dst_network=None,
            src_network=None, exclude_src_network=None,
            dst_port=None, exclude_dst_port=None,
            src_port=None, exclude_src_port=None,
            is_ipv6=False,
            is_change_shaping_rule=False, is_add_shaping_rule=False,
            is_enable_iptables=True,
            shaping_algorithm=None,
            tc_command_output=TcCommandOutput.NOT_SET,
    ):
        self.__device = device

        self.__direction = direction
        self.__bandwidth_rate = bandwidth_rate
        self.__latency_time = HumanReadableTime(latency_time)
        self.__latency_distro_time = HumanReadableTime(latency_distro_time)
        self.__packet_loss_rate = packet_loss_rate  # [%]
        self.__packet_duplicate_rate = packet_duplicate_rate  # [%]
        self.__corruption_rate = corruption_rate  # [%]
        self.__reordering_rate = reordering_rate
        self.__dst_network = dst_network
        self.__exclude_dst_network = exclude_dst_network
        self.__src_network = src_network
        self.__exclude_src_network = exclude_src_network
        self.__src_port = src_port
        self.__exclude_src_port = exclude_src_port
        self.__dst_port = dst_port
        self.__exclude_dst_port = exclude_dst_port
        self.__is_ipv6 = is_ipv6
        self.__is_change_shaping_rule = is_change_shaping_rule
        self.__is_add_shaping_rule = is_add_shaping_rule
        self.__is_enable_iptables = is_enable_iptables
        self.__tc_command_output = tc_command_output

        self.__qdisc_major_id = self.__get_device_qdisc_major_id()

        self.__iptables_ctrl = IptablesMangleController(
            is_enable_iptables, self.ip_version)

        self.__init_shaper(shaping_algorithm)

    def validate(self):
        verify_network_interface(self.device)
        self.__validate_netem_parameter()
        self.__validate_src_network()
        self.__validate_port()

    def __validate_src_network(self):
        if any([
                typepy.is_null_string(self.src_network),
                self.__shaper.algorithm_name == ShapingAlgorithm.HTB,
        ]):
            return

        if not self.is_enable_iptables:
            raise InvalidParameterError(
                "--iptables option required to use --src-network option",
                value=self.is_enable_iptables)

    def validate_bandwidth_rate(self):
        if typepy.is_null_string(self.__bandwidth_rate):
            return

        # convert bandwidth string [K/M/G bit per second] to a number
        bandwidth_rate = Humanreadable(
            self.__bandwidth_rate, kilo_size=KILO_SIZE).to_kilo_bit()

        if not RealNumber(bandwidth_rate).is_type():
            raise InvalidParameterError(
                "bandwidth_rate must be a number", value=bandwidth_rate)

        if bandwidth_rate <= 0:
            raise InvalidParameterError(
                "bandwidth_rate must be greater than zero",
                value=bandwidth_rate)

        no_limit_kbits = get_no_limit_kbits(self.get_tc_device())
        if bandwidth_rate > no_limit_kbits:
            raise InvalidParameterError(
                "bandwidth_rate must be less than {}".format(no_limit_kbits),
                value=bandwidth_rate)

    def sanitize(self):
        self.__dst_network = sanitize_network(
            self.dst_network, self.ip_version)
        self.__src_network = sanitize_network(
            self.src_network, self.ip_version)

    def get_tc_device(self):
        """
        Return a device name that associated network communication direction.
        """

        if self.direction == TrafficDirection.OUTGOING:
            return self.device

        if self.direction == TrafficDirection.INCOMING:
            return self.ifb_device

        raise InvalidParameterError(
            "unknown direction",
            expected=TrafficDirection.LIST, value=self.direction)

    def get_tc_command(self, sub_command):
        valid_sub_command_list = (
            Tc.Subcommand.CLASS, Tc.Subcommand.FILTER, Tc.Subcommand.QDISC)
        if sub_command not in valid_sub_command_list:
            raise InvalidParameterError(
                "unknown sub command",
                expected=valid_sub_command_list, value=sub_command)

        return "tc {:s} {:s}".format(
            sub_command, "change" if self.is_change_shaping_rule else "add")

    def get_command_history(self):
        def tc_filter(command):
            if command == LIST_MANGLE_TABLE_COMMAND:
                return False

            if re.search("^tc .* show dev", command):
                return False

            return True

        return filter(tc_filter, spr.SubprocessRunner.get_history())

    def set_tc(self):
        self.__setup_ifb()

        return self.__shaper.set_shaping()

    def delete_all_tc(self):
        result_list = []

        with logging_context("delete qdisc"):
            returncode = run_command_helper(
                "tc qdisc del dev {:s} root".format(self.device),
                re.compile("RTNETLINK answers: No such file or directory"),
                "no qdisc to delete for the outgoing device.")
            result_list.append(returncode == 0)

        with logging_context("delete ingress qdisc"):
            returncode = run_command_helper(
                "tc qdisc del dev {:s} ingress".format(self.device),
                re.compile("|".join([
                    "RTNETLINK answers: Invalid argument",
                    "RTNETLINK answers: No such file or directory",
                ])),
                "no qdisc to delete for the incoming device.")
            result_list.append(returncode == 0)

        with logging_context("delete ifb device"):
            try:
                result_list.append(self.__delete_ifb_device() == 0)
            except NetworkInterfaceNotFoundError as e:
                logger.debug(e)
                result_list.append(False)

        with logging_context("delete iptables mangle table entries"):
            self.iptables_ctrl.clear()
            # IptablesMangleController.clear()

        return any(result_list)

    def delete_tc(self):
        """
        Delete a specific shaping rule.
        """

        from ._shaping_rule_finder import TcShapingRuleFinder

        rule_finder = TcShapingRuleFinder(logger=logger, tc=self)

        parent = rule_finder.find_parent()
        filter_param = rule_finder.find_filter_param()

        logger.debug("delete_tc: pram={}".format(filter_param))

        if not filter_param:
            message = "shaping rule not found."
            if rule_finder.is_empty_filter_condition():
                message += (
                    " you can delete all of the shaping rules "
                    "with --all option.")
            logger.error(message)

            return 1

        filter_del_command = (
            "tc filter del dev {dev:s} protocol {protocol:s} "
            "parent {parent:} handle {handle:s} prio {prio:} u32".format(
                dev=rule_finder.get_parsed_device(),
                protocol=filter_param.get(Tc.Param.PROTOCOL),
                parent="{:s}:".format(parent.split(":")[0]),
                handle=filter_param.get(Tc.Param.FILTER_ID),
                prio=filter_param.get(Tc.Param.PRIORITY),
            ))

        result = run_command_helper(
            command=filter_del_command, error_regexp=None, message=None)

        rule_finder.clear()
        if not rule_finder.is_any_filter():
            logger.debug("there are no filters remain. delete qdiscs.")
            self.delete_all_tc()

        return result

    def __init_shaper(self, shaping_algorithm):
        if shaping_algorithm is None:
            self.__shaper = None
            return

        if shaping_algorithm == ShapingAlgorithm.HTB:
            self.__shaper = HtbShaper(self)
            return

        if shaping_algorithm == ShapingAlgorithm.TBF:
            self.__shaper = TbfShaper(self)
            return

        raise InvalidParameterError(
            "unknown shaping algorithm",
            expected=ShapingAlgorithm.LIST, value=shaping_algorithm)

    def __validate_network_delay(self):
        try:
            self.latency_time.validate(
                min_value=Tc.ValueRange.LatencyTime.MIN,
                max_value=Tc.ValueRange.LatencyTime.MAX)
        except InvalidParameterError as e:
            raise InvalidParameterError("delay {}".format(str(e)))

        try:
            self.latency_distro_time.validate(
                min_value=Tc.ValueRange.LatencyTime.MIN,
                max_value=Tc.ValueRange.LatencyTime.MAX)
        except InvalidParameterError as e:
            raise InvalidParameterError("delay-distro {}".format(str(e)))

    def __validate_packet_loss_rate(self):
        _validate_within_min_max(
            "loss (packet loss rate)", self.packet_loss_rate,
            self.MIN_PACKET_LOSS_RATE, self.MAX_PACKET_LOSS_RATE, unit="%")

    def __validate_packet_duplicate_rate(self):
        _validate_within_min_max(
            "duplicate (packet duplicate rate)", self.packet_duplicate_rate,
            self.MIN_PACKET_DUPLICATE_RATE, self.MAX_PACKET_DUPLICATE_RATE,
            unit="%")

    def __validate_corruption_rate(self):
        _validate_within_min_max(
            "corruption (packet corruption rate)", self.corruption_rate,
            self.MIN_CORRUPTION_RATE, self.MAX_CORRUPTION_RATE, unit="%")

    def __validate_reordering_rate(self):
        _validate_within_min_max(
            "reordering (packet reordering rate)", self.reordering_rate,
            self.MIN_REORDERING_RATE, self.MAX_REORDERING_RATE, unit="%")

    def __validate_reordering_and_delay(self):
        if self.reordering_rate and not self.latency_time:
            raise InvalidParameterError(
                'reordering needs latency to be specified: set latency > 0')

    def __validate_netem_parameter(self):
        self.validate_bandwidth_rate()
        self.__validate_network_delay()
        self.__validate_packet_loss_rate()
        self.__validate_packet_duplicate_rate()
        self.__validate_corruption_rate()
        self.__validate_reordering_rate()
        self.__validate_reordering_and_delay()

        netem_param_value_list = [
            self.bandwidth_rate,
            self.packet_loss_rate,
            self.packet_duplicate_rate,
            self.corruption_rate,
            self.reordering_rate,
        ]

        if all([
                not RealNumber(netem_param_value).is_type()
                or netem_param_value <= 0
                for netem_param_value in netem_param_value_list
        ] + [
            self.latency_time <= HumanReadableTime(
                Tc.ValueRange.LatencyTime.MIN)
        ]):
            raise InvalidParameterError(
                "there is no valid net emulation parameter value. "
                "at least one or more following parameters are required: "
                "--rate, --delay, --loss, --duplicate, --corrupt, --reordering"
            )

    def __validate_port(self):
        _validate_within_min_max(
            "src_port", self.src_port, self.__MIN_PORT, self.__MAX_PORT,
            unit=None)

        _validate_within_min_max(
            "dst_port", self.dst_port, self.__MIN_PORT, self.__MAX_PORT,
            unit=None)

    def __get_device_qdisc_major_id(self):
        import hashlib
        base_device_hash = hashlib.md5(six.b(self.device)).hexdigest()[:3]
        device_hash_prefix = "1"

        return int(device_hash_prefix + base_device_hash, 16)

    def __setup_ifb(self):
        if self.direction != TrafficDirection.INCOMING:
            return 0

        if typepy.is_null_string(self.ifb_device):
            return -1

        return_code = 0

        return_code |= spr.SubprocessRunner("modprobe ifb").run()

        if self.is_add_shaping_rule or self.is_change_shaping_rule:
            notice_message = None
        else:
            notice_message = self.EXISTS_MSG_TEMPLATE.format(
                "failed to add ip link: ip link already exists.")
        return_code |= run_command_helper(
            "ip link add {:s} type ifb".format(self.ifb_device),
            self.REGEXP_FILE_EXISTS, notice_message)

        return_code |= spr.SubprocessRunner(
            "ip link set dev {:s} up".format(self.ifb_device)).run()

        base_command = "tc qdisc add"
        if self.is_add_shaping_rule or self.is_change_shaping_rule:
            notice_message = None
        else:
            notice_message = self.EXISTS_MSG_TEMPLATE.format(
                "failed to '{:s}': ingress qdisc already exists.".format(
                    base_command))
        return_code |= run_command_helper(
            "{:s} dev {:s} ingress".format(base_command, self.device),
            self.REGEXP_FILE_EXISTS, notice_message)

        return_code |= spr.SubprocessRunner(" ".join([
            "tc filter add",
            "dev {:s}".format(self.device),
            "parent ffff: protocol {:s} u32 match u32 0 0".format(
                self.protocol),
            "flowid {:x}:".format(self.__get_device_qdisc_major_id()),
            "action mirred egress redirect",
            "dev {:s}".format(self.ifb_device),
        ])).run()

        return return_code

    def __delete_ifb_device(self):
        verify_network_interface(self.ifb_device)

        command_list = [
            "tc qdisc del dev {:s} root".format(self.ifb_device),
            "ip link set dev {:s} down".format(self.ifb_device),
            "ip link delete {:s} type ifb".format(self.ifb_device),
        ]

        if all([spr.SubprocessRunner(command).run() != 0 for command in
                command_list]):
            return 2

        return 0



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import itertools

import pingparsing
import pytest
from subprocrunner import SubprocessRunner
import typepy

from tcconfig._const import (
    Tc,
    TrafficDirection,
)
from .common import execute_tcdel


WAIT_TIME = 3  # [sec]


@pytest.fixture
def device_option(request):
    return request.config.getoption("--device")


@pytest.fixture
def local_host_option(request):
    return request.config.getoption("--local-host")


@pytest.fixture
def dst_host_option(request):
    return request.config.getoption("--dst-host")


@pytest.fixture
def transmitter():
    transmitter = pingparsing.PingTransmitter()
    transmitter.ping_option = "-f -q"
    transmitter.waittime = WAIT_TIME

    return transmitter


@pytest.fixture
def pingparser():
    return pingparsing.PingParsing()


class Test_tcset_one_network(object):
    """
    Tests in this class are not executable on CI services.
    Execute the following command at the local environment to running tests:

        python setup.py test --addopts "--device=<test device> \n
            --local-host <IP address> --dst-host=<IP address>"

    These tests expected to execute in the following environment:
       - Linux w/ iputils-ping package
       - English locale (for parsing ping output)
    """

    @pytest.mark.parametrize(["shaping_algo", "delay"], [
        [params[0], params[1]]
        for params in itertools.product(["htb"], [100])
    ])
    def test_src_net_uniform_latency(
            self, device_option, local_host_option, dst_host_option,
            transmitter, pingparser, shaping_algo, delay):
        if device_option is None:
            pytest.skip("device option is null")
        if typepy.is_null_string(local_host_option):
            pytest.skip("local host is null")
        if typepy.is_null_string(dst_host_option):
            pytest.skip("destination host is null")

        execute_tcdel(device_option)
        transmitter.destination_host = dst_host_option

        # w/o latency tc ---
        pingparser.parse(transmitter.ping().stdout)
        without_tc_rtt_avg = pingparser.rtt_avg

        # w/ latency tc ---
        command_list = [
            Tc.Command.TCSET,
            "--device {:s}".format(device_option),
            "--src-network {:s}".format(local_host_option),
            "--delay {:d}ms".format(delay),
            "--shaping-algo {:s}".format(shaping_algo),
        ]
        assert SubprocessRunner(" ".join(command_list)).run() == 0

        pingparser.parse(transmitter.ping().stdout)
        with_tc_rtt_avg = pingparser.rtt_avg

        # assertion ---
        rtt_diff = with_tc_rtt_avg - without_tc_rtt_avg

        print("w/o tc rtt: {} ms".format(without_tc_rtt_avg))
        print("w/ tc rtt: {} ms".format(with_tc_rtt_avg))

        assert rtt_diff > (delay / 2)

        # finalize ---
        execute_tcdel(device_option)


class Test_tcset_exclude(object):
    """
    Tests in this class are not executable on CI services.
    Execute the following command at the local environment to running tests:

        python setup.py test --addopts "--device=<test device> \n
            --local-host <IP address> --dst-host=<IP address>"

    These tests expected to execute in the following environment:
       - Linux w/ iputils-ping package
       - English locale (for parsing ping output)
    """

    @pytest.mark.parametrize(["shaping_algo", "delay"], [
        [params[0], params[1]]
        for params in itertools.product(["htb"], [100])
    ])
    def test_src_net_uniform_latency(
            self, device_option, local_host_option, dst_host_option,
            transmitter, pingparser, shaping_algo, delay):
        if device_option is None:
            pytest.skip("device option is null")
        if typepy.is_null_string(local_host_option):
            pytest.skip("local host is null")
        if typepy.is_null_string(dst_host_option):
            pytest.skip("destination host is null")

        execute_tcdel(device_option)
        transmitter.destination_host = dst_host_option

        # w/o latency tc ---
        pingparser.parse(transmitter.ping().stdout)
        without_tc_rtt_avg = pingparser.rtt_avg

        # w/o latency tc (exclude network) ---
        command_list = [
            Tc.Command.TCSET,
            "--device {:s}".format(device_option),
            "--direction {:s}".format(TrafficDirection.INCOMING),
            "--exclude-dst-network {:s}".format(local_host_option),
            "--exclude-src-network {:s}".format(dst_host_option),
            "--delay {:d}ms".format(delay),
            "--shaping-algo {:s}".format(shaping_algo),
        ]
        assert SubprocessRunner(" ".join(command_list)).run() == 0

        pingparser.parse(transmitter.ping().stdout)
        exclude_tc_rtt_avg = pingparser.rtt_avg

        # assertion ---
        rtt_diff = exclude_tc_rtt_avg - without_tc_rtt_avg

        print("w/o tc rtt: {} ms".format(without_tc_rtt_avg))
        print("exclude tc rtt: {} ms".format(exclude_tc_rtt_avg))

        assert rtt_diff < (delay / 10)

        # finalize ---
        execute_tcdel(device_option)



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import


class NetworkInterfaceNotFoundError(Exception):
    """
    Exception raised when network interface not found.
    """


class ModuleNotFoundError(Exception):
    """
    Exception raised when mandatory kernel module not found.
    """


class TcCommandExecutionError(Exception):
    """
    Exception raised when failed to execute a ``tc`` command.
    """


class TcAlreadyExist(TcCommandExecutionError):
    """
    Exception raised when a traffic shaping rule already exist.
    """


class EmptyParameterError(ValueError):
    """
    Exception raised when a parameter value is empty value.
    """


class InvalidParameterError(ValueError):
    """
    Exception raised when an invalid parameter specified for
    a traffic shaping rule.
    """

    def __init__(self, *args, **kwargs):
        self.__value = kwargs.pop("value", None)
        self.__expected = kwargs.pop("expected", None)

        super(ValueError, self).__init__(*args)

    def __str__(self, *args, **kwargs):
        item_list = [ValueError.__str__(self, *args, **kwargs)]
        extra_item_list = []

        if self.__expected:
            extra_item_list.append("expected={}".format(self.__expected))

        if self.__value:
            extra_item_list.append("value={}".format(self.__value))

        if extra_item_list:
            item_list.extend([":", ", ".join(extra_item_list)])

        return " ".join(item_list)

    def __repr__(self, *args, **kwargs):
        return self.__str__(*args, **kwargs)


class UnitNotFoundError(InvalidParameterError):
    """
    """



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from collections import namedtuple
import re

import typepy

from ._error import (
    InvalidParameterError,
    UnitNotFoundError,
)
from ._logger import logger


ByteUnit = namedtuple("ByteUnit", "regexp factor")

_RE_NUMBER = re.compile("^[\-\+]?[0-9\.]+")


class Humanreadable(object):
    __UNIT_LIST = [
        ByteUnit(regexp=re.compile("^b$", re.IGNORECASE), factor=0),
        ByteUnit(regexp=re.compile("^bps$", re.IGNORECASE), factor=0),
        ByteUnit(regexp=re.compile("^k$", re.IGNORECASE), factor=1),
        ByteUnit(regexp=re.compile("^kbps$", re.IGNORECASE), factor=1),
        ByteUnit(regexp=re.compile("^m$", re.IGNORECASE), factor=2),
        ByteUnit(regexp=re.compile("^mbps$", re.IGNORECASE), factor=2),
        ByteUnit(regexp=re.compile("^g$", re.IGNORECASE), factor=3),
        ByteUnit(regexp=re.compile("^gbps$", re.IGNORECASE), factor=3),
        ByteUnit(regexp=re.compile("^t$", re.IGNORECASE), factor=4),
        ByteUnit(regexp=re.compile("^gbps$", re.IGNORECASE), factor=3),
        ByteUnit(regexp=re.compile("^p$", re.IGNORECASE), factor=5),
    ]

    def __init__(self, readable_size, kilo_size=1024):
        """
        String converter that human-readable byte size to a number.

        :param str readable_size: human readable size (bytes). e.g. 256 M
        :param int kilo_size: size of ``kilo``. 1024 or 1000
        """

        self.__readable_size = readable_size
        self.kilo_size = kilo_size

        self.__validate_kilo_size()

    def to_bit(self):
        """
        :raises InvalidParameterError:
        """

        logger.debug("human readable size to bit: {}".format(
            self.__readable_size))

        if not typepy.is_not_null_string(self.__readable_size):
            raise TypeError(
                "readable_size must be a string: actual={}".format(
                    self.__readable_size))

        self.__readable_size = self.__readable_size.strip()

        try:
            size = _RE_NUMBER.search(self.__readable_size).group()
        except AttributeError:
            raise InvalidParameterError(
                "invalid value", value=self.__readable_size)
        size = float(size)
        if size < 0:
            raise InvalidParameterError(
                "size must be greater or equals to zero", value=size)

        unit = _RE_NUMBER.sub("", self.__readable_size).strip().lower()

        return size * self.__get_coefficient(unit)

    def to_kilo_bit(self):
        """
        :param str readable_size: human readable size (bytes). e.g. 256 M
        :raises ValueError:
        """

        return self.to_bit() / self.kilo_size

    def __validate_kilo_size(self):
        VALID_KILLO_SIZE = [1000, 1024]

        if self.kilo_size not in VALID_KILLO_SIZE:
            raise InvalidParameterError(
                "invalid kilo size",
                expected=VALID_KILLO_SIZE, value=self.kilo_size)

    def __get_coefficient(self, unit_str):
        self.__validate_kilo_size()

        for unit in self.__UNIT_LIST:
            if unit.regexp.search(unit_str):
                return self.kilo_size ** unit.factor

        raise UnitNotFoundError("unit not found: value={}".format(unit_str))


class HumanReadableTime(object):
    __VALID_MINUTE_UNIT_LIST = ["m", "min", "mins", "minute", "minutes"]
    __VALID_SEC_UNIT_LIST = ["s", "sec", "secs", "second", "seconds"]
    __VALID_MSEC_UNIT_LIST = [
        "ms", "msec", "msecs",  "millisecond", "milliseconds"]
    __VALID_USEC_UNIT_LIST = [
        "us", "usec", "usecs", "microsecond", "microseconds"]
    __VALID_UNIT_LIST = (
        __VALID_MINUTE_UNIT_LIST + __VALID_SEC_UNIT_LIST +
        __VALID_MSEC_UNIT_LIST + __VALID_USEC_UNIT_LIST)

    def __init__(self, readable_time):
        self.__readable_time = readable_time

    def __eq__(self, other):
        return self.get_msec() == other.get_msec()

    def __ne__(self, other):
        return self.get_msec() != other.get_msec()

    def __lt__(self, other):
        return self.get_msec() < other.get_msec()

    def __le__(self, other):
        return self.get_msec() <= other.get_msec()

    def __gt__(self, other):
        return self.get_msec() > other.get_msec()

    def __ge__(self, other):
        return self.get_msec() >= other.get_msec()

    def __repr__(self):
        return self.get_value()

    @classmethod
    def get_valid_unit_list(cls):
        return cls.__VALID_UNIT_LIST

    def get_value(self):
        self.__preprocess()

        return "{:f}{:s}".format(self.__number, self.__unit)

    def get_msec(self):
        self.__preprocess()

        coef = 1
        if self.__unit == "sec":
            coef = 1000
        elif self.__unit == "us":
            coef = .001

        return self.__number * coef

    def validate(self, min_value=None, max_value=None):
        if min_value is not None:
            if not isinstance(min_value, HumanReadableTime):
                min_value = HumanReadableTime(min_value)

            if self < min_value:
                raise InvalidParameterError(
                    "time value is too low",
                    expected="greater than or equal to {}".format(min_value),
                    value=self)

        if max_value is not None:
            if not isinstance(max_value, HumanReadableTime):
                max_value = HumanReadableTime(max_value)

            if self > max_value:
                raise InvalidParameterError(
                    "time value is too high",
                    expected="less than or equal to {}".format(max_value),
                    value=self)

        if self.__unit not in self.__VALID_UNIT_LIST:
            raise UnitNotFoundError(
                "unknown unit".format(self.__readable_time))

    def __normalize(self):
        if self.__unit in self.__VALID_SEC_UNIT_LIST:
            self.__unit = "sec"
        elif self.__unit in self.__VALID_MSEC_UNIT_LIST:
            self.__unit = "ms"
        elif self.__unit in self.__VALID_USEC_UNIT_LIST:
            self.__unit = "us"
        elif self.__unit in self.__VALID_MINUTE_UNIT_LIST:
            self.__number *= 60
            self.__unit = "sec"

    def __get_number(self):
        match = _RE_NUMBER.search(self.__readable_time)
        if not match:
            raise InvalidParameterError(
                "human-readable time must be include a number",
                value=self.__readable_time)

        return float(match.group())

    def __get_unit(self):
        if typepy.type.RealNumber(self.__readable_time).is_type():
            # if the input value is real numbers consider unit as milliseconds.
            return "ms"

        return _RE_NUMBER.sub("", self.__readable_time).strip().lower()

    def __preprocess(self):
        self.__number = self.__get_number()
        self.__unit = self.__get_unit()
        self.validate()
        self.__normalize()



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import
from __future__ import unicode_literals

import contextlib
import errno
import sys

import logbook
import six
import typepy

import subprocrunner as spr

from ._const import (
    IPV6_OPTION_ERROR_MSG_FORMAT,
    KILO_SIZE,
    Network,
    Tc,
    TcCommandOutput,
)
from ._error import NetworkInterfaceNotFoundError
from ._logger import logger


@contextlib.contextmanager
def logging_context(name):
    logger.debug("|---- {:s}: {:s} -----".format("start", name))
    try:
        yield
    finally:
        logger.debug("----- {:s}: {:s} ----|".format("complete", name))


def check_tc_command_installation():
    try:
        spr.Which("tc").verify()
    except spr.CommandNotFoundError as e:
        logger.error(e)
        sys.exit(errno.ENOENT)


def get_anywhere_network(ip_version):
    ip_version_n = typepy.type.Integer(ip_version).try_convert()

    if ip_version_n == 4:
        return Network.Ipv4.ANYWHERE

    if ip_version_n == 6:
        return Network.Ipv6.ANYWHERE

    raise ValueError("unknown ip version: {}".format(ip_version))


def get_iproute2_upper_limite_rate():
    """
    :return: Upper bandwidth rate limit of iproute2 [Kbps].
    :rtype: int
    """

    from ._converter import Humanreadable

    # upper bandwidth rate limit of iproute2 was 34,359,738,360
    # bits per second older than 3.14.0
    # http://git.kernel.org/cgit/linux/kernel/git/shemminger/iproute2.git/commit/?id=8334bb325d5178483a3063c5f06858b46d993dc7

    return Humanreadable(
        "32G", kilo_size=KILO_SIZE).to_kilo_bit()


def get_no_limit_kbits(tc_device):
    if typepy.is_null_string(tc_device):
        return get_iproute2_upper_limite_rate()

    try:
        speed_value = read_iface_speed(tc_device)
    except IOError:
        return get_iproute2_upper_limite_rate()

    if speed_value < 0:
        # default to the iproute2 upper limit when speed value is -1 in
        # paravirtualized network interfaces
        return get_iproute2_upper_limite_rate()
    return min(
        speed_value * KILO_SIZE,
        get_iproute2_upper_limite_rate())


def initialize(options):
    import subprocrunner
    from ._logger import set_log_level

    set_log_level(options.log_level)
    subprocrunner.SubprocessRunner.is_save_history = True

    if options.is_output_stacktrace:
        subprocrunner.SubprocessRunner.is_output_stacktrace = (
            options.is_output_stacktrace)


def is_anywhere_network(network, ip_version):
    try:
        network = network.strip()
    except AttributeError as e:
        raise ValueError(e)

    if ip_version == 4:
        return network == get_anywhere_network(ip_version)

    if ip_version == 6:
        return network in (
            get_anywhere_network(ip_version), "0:0:0:0:0:0:0:0/0")

    raise ValueError("invalid ip version: {}".format(ip_version))


def is_execute_tc_command(tc_command_output):
    return tc_command_output == TcCommandOutput.NOT_SET


def normalize_tc_value(tc_obj):
    import ipaddress

    try:
        tc_obj.sanitize()
    except ipaddress.AddressValueError as e:
        logger.error(IPV6_OPTION_ERROR_MSG_FORMAT.format(e))
        sys.exit(errno.EINVAL)
    except ValueError as e:
        logger.error(e)
        sys.exit(errno.EINVAL)


def read_iface_speed(tc_device):
    with open("/sys/class/net/{:s}/speed".format(tc_device)) as f:
        return int(f.read().strip())


def run_command_helper(command, error_regexp, message, exception=None):
    if logger.level != logbook.DEBUG:
        spr.set_logger(is_enable=False)

    proc = spr.SubprocessRunner(command)
    proc.run()

    if logger.level != logbook.DEBUG:
        spr.set_logger(is_enable=True)

    if proc.returncode == 0:
        return 0

    match = error_regexp.search(proc.stderr)
    if match is None:
        logger.error(proc.stderr)
        return proc.returncode

    if typepy.is_not_null_string(message):
        logger.notice(message)

    if exception is not None:
        raise exception(command)

    return proc.returncode


def run_tc_show(subcommand, device):
    if subcommand not in Tc.Subcommand.LIST:
        raise ValueError("unexpected tc sub command: {}".format(subcommand))

    verify_network_interface(device)

    runner = spr.SubprocessRunner(
        "tc {:s} show dev {:s}".format(subcommand, device))
    runner.run()

    return runner.stdout


def sanitize_network(network, ip_version):
    """
    :return: Network string
    :rtype: str
    :raises ValueError: if the network string is invalid.
    """

    import ipaddress

    if typepy.is_null_string(network) or network.lower() == "anywhere":
        return get_anywhere_network(ip_version)

    try:
        if ip_version == 4:
            ipaddress.IPv4Address(network)
            return network + "/32"

        if ip_version == 6:
            return ipaddress.IPv6Address(network).compressed
    except ipaddress.AddressValueError:
        pass

    # validate network str ---

    if ip_version == 4:
        return ipaddress.IPv4Network(six.text_type(network)).compressed

    if ip_version == 6:
        return ipaddress.IPv6Network(six.text_type(network)).compressed

    raise ValueError("unexpected ip version: {}".format(ip_version))


def verify_network_interface(device):
    try:
        import netifaces
    except ImportError:
        return

    if device not in netifaces.interfaces():
        raise NetworkInterfaceNotFoundError(
            "network interface not found: {}".format(device))


def write_tc_script(tcconfig_command, command_history, filename_suffix=None):
    import datetime
    import io
    import os

    filename_item_list = [tcconfig_command]
    if typepy.is_not_null_string(filename_suffix):
        filename_item_list.append(filename_suffix)

    script_line_list = [
        "#!/bin/sh",
        "",
        "# tc script file:",
    ]

    if tcconfig_command != Tc.Command.TCSHOW:
        script_line_list.extend([
            "#   the following command sequence lead to equivalent results as",
            "#   '{:s}'.".format(
                _get_original_tcconfig_command(tcconfig_command)),
        ])

    script_line_list.extend([
        "#   created by {:s} on {:s}.".format(
            tcconfig_command,
            datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S%z")),
        "",
        command_history,
    ])

    filename = "_".join(filename_item_list) + ".sh"
    with io.open(filename, "w", encoding="utf8") as fp:
        fp.write("\n".join(script_line_list) + "\n")

    os.chmod(filename, 0o755)
    logger.info("written a tc script to '{:s}'".format(filename))


def _get_original_tcconfig_command(tcconfig_command):
    return " ".join([tcconfig_command] + [
        command_item for command_item in sys.argv[1:]
        if command_item != "--tc-script"
    ])



#!/usr/bin/env python
# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import
from __future__ import print_function

import errno
import ipaddress
import sys

import logbook
import six
import subprocrunner
import typepy

import pyparsing as pp

from ._argparse_wrapper import ArgparseWrapper
from ._common import (
    check_tc_command_installation,
    initialize,
    is_execute_tc_command,
    normalize_tc_value,
    write_tc_script,
)
from ._const import (
    VERSION,
    IPV6_OPTION_ERROR_MSG_FORMAT,
    Network,
    ShapingAlgorithm,
    Tc,
    TcCommandOutput,
    TrafficDirection,
)
from ._converter import HumanReadableTime
from ._error import (
    InvalidParameterError,
    ModuleNotFoundError,
    NetworkInterfaceNotFoundError,
    UnitNotFoundError,
)
from ._logger import (
    LOG_FORMAT_STRING,
    logger,
    set_log_level,
)
from ._shaping_rule_finder import TcShapingRuleFinder
from .traffic_control import TrafficControl


logbook.StderrHandler(
    level=logbook.DEBUG, format_string=LOG_FORMAT_STRING).push_application()


def get_arg_parser():
    parser = ArgparseWrapper(VERSION)

    group = parser.parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "-d", "--device",
        help="network device name (e.g. eth0)")
    group.add_argument(
        "-f", "--config-file",
        help="""setting traffic controls from a configuration file.
        output file of the tcshow.
        """)

    group = parser.parser.add_mutually_exclusive_group()
    group.add_argument(
        "--overwrite", action="store_true", default=False,
        help="overwrite existing traffic shaping rules.")
    group.add_argument(
        "--change", dest="is_change_shaping_rule", action="store_true",
        default=False,
        help="""change existing traffic shaping rules to the new one.
        this option reduces the shaping rule switching side effect
        (such as traffic spike) compared to --overwrite option.
        note: the tcset command fail when there are no existing shaping rules.
        """)
    group.add_argument(
        "--add", dest="is_add_shaping_rule", action="store_true",
        default=False,
        help="add a traffic shaping rule in addition to existing rules.")

    group = parser.parser.add_argument_group("Traffic Control Parameters")
    group.add_argument(
        "--rate", "--bandwidth-rate", dest="bandwidth_rate",
        help="""network bandwidth rate [bit per second].
        valid units are either: K/M/G/Kbps/Mbps/Gbps
        e.g. --rate 10Mbps
        """)
    group.add_argument(
        "--delay", dest="network_latency",
        default=Tc.ValueRange.LatencyTime.MIN,
        help="""round trip network delay. the valid range is
        from {min_value:} to {max_value:}.
        valid time units are: {unit}.
        if no unit string found, considered milliseconds as the time unit.
        (default=%(default)s)
        """.format(
            min_value=Tc.ValueRange.LatencyTime.MIN,
            max_value=Tc.ValueRange.LatencyTime.MAX,
            unit="/".join(HumanReadableTime.get_valid_unit_list())))
    group.add_argument(
        "--delay-distro", dest="latency_distro_time",
        default=Tc.ValueRange.LatencyTime.MIN,
        help="""distribution of network latency becomes X +- Y
        (normal distribution). Here X is the value of --delay option and
        Y is the value of --delay-dist option).
        network latency distribution is uniform, without this option.
        valid time units are: {unit}.
        if no unit string found, considered milliseconds as the time unit.
        """.format(unit="/".join(HumanReadableTime.get_valid_unit_list())))
    group.add_argument(
        "--loss", dest="packet_loss_rate", type=float, default=0,
        help="""round trip packet loss rate [%%]. the valid range is from {:d}
        to {:d}. (default=%(default)s)
        """.format(
            TrafficControl.MIN_PACKET_LOSS_RATE,
            TrafficControl.MAX_PACKET_LOSS_RATE))
    group.add_argument(
        "--duplicate", dest="packet_duplicate_rate", type=float, default=0,
        help="""round trip packet duplicate rate [%%]. the valid range is
        from {:d} to {:d}. (default=%(default)s)
        """.format(
            TrafficControl.MIN_PACKET_DUPLICATE_RATE,
            TrafficControl.MAX_PACKET_DUPLICATE_RATE))
    group.add_argument(
        "--corrupt", dest="corruption_rate", type=float, default=0,
        help="""packet corruption rate [%%]. the valid range is from {:d}
        to {:d}. packet corruption means single bit error at a random offset in
        the packet. (default=%(default)s)
        """.format(
            TrafficControl.MIN_CORRUPTION_RATE,
            TrafficControl.MAX_CORRUPTION_RATE))
    group.add_argument(
        "--reordering", dest="reordering_rate", type=float, default=0,
        help="""packet reordering rate [%%]. the valid range is from {:d}
        to {:d}. (default=%(default)s)
        """.format(
            TrafficControl.MIN_REORDERING_RATE,
            TrafficControl.MAX_REORDERING_RATE))
    group.add_argument(
        "--shaping-algo", dest="shaping_algorithm",
        choices=[ShapingAlgorithm.HTB, ShapingAlgorithm.HTB],
        default=ShapingAlgorithm.HTB,
        help="shaping algorithm. defaults to %(default)s (recommended).")
    group.add_argument(
        "--iptables", dest="is_enable_iptables",
        action="store_true", default=False,
        help="use iptables to traffic control.")

    group = parser.add_routing_group()
    group.add_argument(
        "--exclude-dst-network",
        help="""exclude a shaping rule for a specific destination
        IP-address/network.
        """)
    group.add_argument(
        "--exclude-src-network",
        help="""exclude a shaping rule for a specific source
        IP-address/network.
        """)
    group.add_argument(
        "--exclude-dst-port",
        help="exclude a shaping rule for a specific destination port.")
    group.add_argument(
        "--exclude-src-port",
        help="exclude a shaping rule for a specific source port.")

    return parser.parser


def verify_netem_module():
    import re

    runner = subprocrunner.SubprocessRunner("lsmod")
    if runner.run() != 0:
        raise OSError("failed to execute lsmod")

    if re.search(r"\bsch_netem\b", runner.stdout) is None:
        raise ModuleNotFoundError("sch_netem module not found")


class TcConfigLoader(object):
    def __init__(self, logger):
        self.__logger = logger
        self.__config_table = None
        self.is_overwrite = False

    def load_tcconfig(self, config_file_path):
        import json
        from voluptuous import Schema, Required, Any, ALLOW_EXTRA

        schema = Schema({
            Required(six.text_type): {
                Any(*TrafficDirection.LIST): {
                    six.text_type: {
                        six.text_type: Any(six.text_type, int, float)
                    },
                }
            },
        }, extra=ALLOW_EXTRA)

        with open(config_file_path) as fp:
            self.__config_table = json.load(fp)

        schema(self.__config_table)
        self.__logger.debug("tc config file: {:s}".format(
            json.dumps(self.__config_table, indent=4)))

    def get_tcconfig_command_list(self):
        command_list = []

        for device, device_table in six.iteritems(self.__config_table):
            device_option = "--device={:s}".format(device)

            if self.is_overwrite:
                command_list.append("{:s} {:s}".format(
                    Tc.Command.TCDEL, device_option))

            for direction, direction_table in six.iteritems(device_table):
                is_first_set = True

                for tc_filter, filter_table in six.iteritems(direction_table):
                    self.__logger.debug(
                        "is_first_set={}, filter='{}', table={}".format(
                            is_first_set, tc_filter, filter_table))

                    if not filter_table:
                        continue

                    option_list = [
                        device_option,
                        "--direction={:s}".format(direction),
                    ]
                    for key, value in six.iteritems(filter_table):
                        arg_item = "--{:s}={}".format(key, value)

                        parse_result = get_arg_parser().parse_known_args(
                            ["--device", "dummy", arg_item])
                        if parse_result[1]:
                            self.__logger.debug(
                                "unknown parameter: key={}, value={}".format(
                                    key, value))
                            continue

                        option_list.append(arg_item)

                    try:
                        dst_network = self.__parse_tc_filter_network(tc_filter)
                        if dst_network not in (
                                Network.Ipv4.ANYWHERE, Network.Ipv6.ANYWHERE):
                            option_list.append(
                                "--dst-network={:s}".format(dst_network))
                    except pp.ParseException:
                        pass

                    try:
                        src_port = self.__parse_tc_filter_src_port(tc_filter)
                        option_list.append("--src-port={}".format(src_port))
                    except pp.ParseException:
                        pass

                    try:
                        dst_port = self.__parse_tc_filter_dst_port(tc_filter)
                        option_list.append("--dst-port={}".format(dst_port))
                    except pp.ParseException:
                        pass

                    if not is_first_set:
                        option_list.append("--add")

                    is_first_set = False

                    command_list.append(
                        " ".join([Tc.Command.TCSET] + option_list))

        return command_list

    @staticmethod
    def __parse_tc_filter_network(text):
        network_pattern = (
            pp.SkipTo("{:s}=".format(Tc.Param.DST_NETWORK), include=True) +
            pp.Word(pp.alphanums + "." + "/"))

        return network_pattern.parseString(text)[-1]

    @staticmethod
    def __parse_tc_filter_src_port(text):
        port_pattern = (
            pp.SkipTo("{:s}=".format(Tc.Param.SRC_PORT), include=True) +
            pp.Word(pp.nums))

        return port_pattern.parseString(text)[-1]

    @staticmethod
    def __parse_tc_filter_dst_port(text):
        port_pattern = (
            pp.SkipTo("{:s}=".format(Tc.Param.DST_PORT), include=True) +
            pp.Word(pp.nums))

        return port_pattern.parseString(text)[-1]


def set_tc_from_file(logger, config_file_path, is_overwrite):
    return_code = 0

    loader = TcConfigLoader(logger)
    loader.is_overwrite = is_overwrite

    try:
        loader.load_tcconfig(config_file_path)
    except IOError as e:
        logger.error(e)
        return errno.EIO

    for tcconfig_command in loader.get_tcconfig_command_list():
        return_code |= subprocrunner.SubprocessRunner(
            tcconfig_command).run()

    return return_code


def main():
    options = get_arg_parser().parse_args()

    initialize(options)

    if is_execute_tc_command(options.tc_command_output):
        check_tc_command_installation()
    else:
        subprocrunner.SubprocessRunner.default_is_dry_run = True

    try:
        verify_netem_module()
    except ModuleNotFoundError as e:
        logger.debug(str(e))
    except subprocrunner.CommandNotFoundError as e:
        logger.error(str(e))

    if typepy.is_not_null_string(options.config_file):
        return set_tc_from_file(logger, options.config_file, options.overwrite)

    subprocrunner.SubprocessRunner.clear_history()

    try:
        tc = TrafficControl(
            options.device,
            direction=options.direction,
            bandwidth_rate=options.bandwidth_rate,
            latency_time=options.network_latency,
            latency_distro_time=options.latency_distro_time,
            packet_loss_rate=options.packet_loss_rate,
            packet_duplicate_rate=options.packet_duplicate_rate,
            corruption_rate=options.corruption_rate,
            reordering_rate=options.reordering_rate,
            dst_network=options.dst_network,
            exclude_dst_network=options.exclude_dst_network,
            src_network=options.src_network,
            exclude_src_network=options.exclude_src_network,
            src_port=options.src_port,
            exclude_src_port=options.exclude_src_port,
            dst_port=options.dst_port,
            exclude_dst_port=options.exclude_dst_port,
            is_ipv6=options.is_ipv6,
            is_change_shaping_rule=options.is_change_shaping_rule,
            is_add_shaping_rule=options.is_add_shaping_rule,
            is_enable_iptables=options.is_enable_iptables,
            shaping_algorithm=options.shaping_algorithm,
            tc_command_output=options.tc_command_output,
        )
    except UnitNotFoundError:
        logger.error(
            "--rate/--bandwidth-rate require an unit: K/M/Kbps/Mbps/etc.")
        return errno.EINVAL

    try:
        tc.validate()
    except (NetworkInterfaceNotFoundError) as e:
        logger.error(str(e))
        return errno.EINVAL
    except ipaddress.AddressValueError as e:
        logger.error(IPV6_OPTION_ERROR_MSG_FORMAT.format(e))
        return errno.EINVAL
    except InvalidParameterError as e:
        logger.error(e)
        return errno.EINVAL

    normalize_tc_value(tc)

    if options.overwrite:
        if options.log_level == logbook.INFO:
            set_log_level(logbook.ERROR)

        try:
            tc.delete_all_tc()
        except NetworkInterfaceNotFoundError:
            pass

        set_log_level(options.log_level)

    if (
            options.is_add_shaping_rule and
            TcShapingRuleFinder(logger=logger, tc=tc).is_exist_rule()
    ):
        logger.error(
            "adding a shaping rule failed. a shaping rule for the same "
            "network/port already exist. try to execute with: "
            "(a) --overwrite option if you want to overwrite "
            "the existing rules. "
            "(b) --change option if you want to change the existing "
            "rule parameter.")
        return errno.EINVAL

    return_code = tc.set_tc()
    command_history = "\n".join(tc.get_command_history())

    if options.tc_command_output == TcCommandOutput.STDOUT:
        print(command_history)
        return 0

    if options.tc_command_output == TcCommandOutput.SCRIPT:
        write_tc_script(
            Tc.Command.TCSET, command_history, filename_suffix=options.device)
        return 0

    logger.debug("command history\n{}".format(command_history))

    return return_code


if __name__ == '__main__':
    sys.exit(main())



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import
from __future__ import unicode_literals

from collections import OrderedDict
import ipaddress
import json
import re

import typepy

import pyparsing as pp

from .._common import sanitize_network
from .._const import Tc
from .._logger import logger
from ._interface import AbstractParser


class TcFilterParser(AbstractParser):

    class FilterMatchIdIpv4(object):
        INCOMING_NETWORK = 12
        OUTGOING_NETWORK = 16
        PORT = 20

    class FilterMatchIdIpv6(object):
        INCOMING_NETWORK_LIST = [8, 12, 16, 20]
        OUTGOING_NETWORK_LIST = [24, 28, 32, 36]
        PORT = 40

    __FILTER_FLOWID_PATTERN = (
        pp.Literal("filter parent") +
        pp.SkipTo("flowid", include=True) +
        pp.Word(pp.hexnums + ":")
    )
    __FILTER_PROTOCOL_PATTERN = (
        pp.Literal("filter parent") +
        pp.SkipTo("protocol", include=True) +
        pp.Word(pp.alphanums)
    )
    __FILTER_PRIORITY_PATTERN = (
        pp.Literal("filter parent") +
        pp.SkipTo("pref", include=True) +
        pp.Word(pp.nums))
    __FILTER_ID_PATTERN = (
        pp.Literal("filter parent") +
        pp.SkipTo("fh", include=True) +
        pp.Word(pp.hexnums + ":")
    )
    __FILTER_MATCH_PATTERN = (
        pp.Literal("match") +
        pp.Word(pp.alphanums + "/") +
        pp.Literal("at") +
        pp.Word(pp.nums)
    )
    __FILTER_MANGLE_MARK_PATTERN = (
        pp.Literal("filter parent") +
        pp.SkipTo("handle", include=True) +
        pp.Word(pp.hexnums) +
        pp.SkipTo("classid", include=True) +
        pp.Word(pp.hexnums + ":")
    )

    @property
    def protocol(self):
        return self.__protocol

    @property
    def _tc_subcommand(self):
        return Tc.Subcommand.FILTER

    def __init__(self, con, ip_version):
        super(TcFilterParser, self).__init__()

        self.__con = con
        self.__ip_version = ip_version
        self.__buffer = None
        self.__parse_idx = 0

        self.__protocol = None

        self._clear()

    def parse(self, device, text):
        self._clear()

        if typepy.is_null_string(text):
            return []

        filter_data_matrix = []
        self.__buffer = self._to_unicode(text).splitlines()
        self.__parse_idx = 0

        while self.__parse_idx < len(self.__buffer):
            line = self._to_unicode(self.__buffer[self.__parse_idx].strip())
            self.__parse_idx += 1

            if typepy.is_null_string(line):
                continue

            self.__device = device

            try:
                self.__parse_mangle_mark(line)
            except pp.ParseException:
                logger.debug("failed to parse mangle: {}".format(line))
            else:
                filter_data_matrix.append({
                    Tc.Param.DEVICE: self.__device,
                    Tc.Param.CLASS_ID: self.__classid,
                    Tc.Param.HANDLE: self.__handle,
                })
                self._clear()
                continue

            tc_filter = self.__get_filter()

            try:
                self.__parse_flow_id(line)
                self.__parse_protocol(line)
                self.__parse_priority(line)
                self.__parse_filter_id(line)

                if tc_filter.get(Tc.Param.FLOW_ID):
                    logger.debug("store filter: {}".format(tc_filter))
                    filter_data_matrix.append(tc_filter)
                    self._clear()

                    self.__parse_flow_id(line)
                    self.__parse_protocol(line)
                    self.__parse_priority(line)
                    self.__parse_filter_id(line)

                continue
            except pp.ParseException:
                logger.debug("failed to parse flow id: {}".format(line))

            try:
                if self.__ip_version == 4:
                    self.__parse_filter_ipv4(line)
                elif self.__ip_version == 6:
                    self.__parse_filter_ipv6(line)
                else:
                    raise ValueError(
                        "unknown ip version: {}".format(self.__ip_version))
            except pp.ParseException:
                logger.debug("failed to parse filter: {}".format(line))

        if self.__flow_id:
            filter_data_matrix.append(self.__get_filter())

        if filter_data_matrix:
            self.__con.create_table_from_data_matrix(
                table_name=self._tc_subcommand,
                attr_name_list=list(self.__get_filter()),
                data_matrix=filter_data_matrix)

        logger.debug("tc {:s} parse result: {}".format(
            self._tc_subcommand, json.dumps(filter_data_matrix, indent=4)))

        return filter_data_matrix

    def parse_incoming_device(self, text):
        if typepy.is_null_string(text):
            return None

        match = re.search(
            "Egress Redirect to device ifb[\d]+",
            self._to_unicode(text), re.MULTILINE)
        if match is None:
            return None

        return re.search("ifb[\d]+", match.group()).group()

    def _clear(self):
        self.__device = None
        self.__filter_id = None
        self.__flow_id = None
        self.__protocol = None
        self.__priority = None
        self.__filter_src_network = None
        self.__filter_dst_network = None
        self.__filter_src_port = None
        self.__filter_dst_port = None

        self.__handle = None
        self.__classid = None

    def __get_filter(self):
        tc_filter = OrderedDict()
        tc_filter[Tc.Param.DEVICE] = self.__device
        tc_filter[Tc.Param.FILTER_ID] = self.__filter_id
        tc_filter[Tc.Param.FLOW_ID] = self.__flow_id
        tc_filter[Tc.Param.PROTOCOL] = self.protocol
        tc_filter[Tc.Param.PRIORITY] = self.__priority
        tc_filter[Tc.Param.SRC_NETWORK] = sanitize_network(
            self.__filter_src_network, self.__ip_version)
        tc_filter[Tc.Param.DST_NETWORK] = sanitize_network(
            self.__filter_dst_network, self.__ip_version)
        tc_filter[Tc.Param.SRC_PORT] = self.__filter_src_port
        tc_filter[Tc.Param.DST_PORT] = self.__filter_dst_port

        return tc_filter

    def __parse_flow_id(self, line):
        parsed_list = self.__FILTER_FLOWID_PATTERN.parseString(line)
        self.__flow_id = parsed_list[-1]
        logger.debug("succeed to parse flow id: flow-id={}, line={}".format(
            self.__flow_id, line))

    def __parse_protocol(self, line):
        parsed_list = self.__FILTER_PROTOCOL_PATTERN.parseString(line)
        self.__protocol = parsed_list[-1]
        logger.debug("succeed to parse protocol: protocol={}, line={}".format(
            self.__protocol, line))

    def __parse_priority(self, line):
        parsed_list = self.__FILTER_PRIORITY_PATTERN.parseString(line)
        self.__priority = int(parsed_list[-1])
        logger.debug("succeed to parse priority: priority={}, line={}".format(
            self.__priority, line))

    def __parse_filter_id(self, line):
        parsed_list = self.__FILTER_ID_PATTERN.parseString(line)
        self.__filter_id = parsed_list[-1]
        logger.debug("succeed to parse filter id: filter-id={}, line={}".format(
            self.__filter_id, line))

    def __parse_mangle_mark(self, line):
        parsed_list = self.__FILTER_MANGLE_MARK_PATTERN.parseString(line)
        self.__classid = parsed_list[-1]
        self.__handle = int("0" + parsed_list[-3], 16)
        logger.debug(
            "succeed to parse mangle mark: "
            "classid={}, handle={}, line={}".format(
                self.__classid, self.__handle, line))

    def __parse_filter_ip_line(self, line):
        parsed_list = self.__FILTER_MATCH_PATTERN.parseString(line)
        value_hex, mask_hex = parsed_list[1].split("/")
        match_id = int(parsed_list[3])

        return (value_hex, mask_hex, match_id)

    def __parse_filter_ipv4_network(self, value_hex, mask_hex, match_id):
        ipaddr = ".".join([
            str(int(value_hex[i: i + 2], 16))
            for i in range(0, len(value_hex), 2)
        ])
        netmask = bin(int(mask_hex, 16)).count("1")
        network = "{:s}/{:d}".format(ipaddr, netmask)

        if match_id == self.FilterMatchIdIpv4.INCOMING_NETWORK:
            self.__filter_src_network = network
        elif match_id == self.FilterMatchIdIpv4.OUTGOING_NETWORK:
            self.__filter_dst_network = network
        else:
            logger.warn("unknown match id: {}".format(match_id))

    def __parse_filter_ipv6_network(self, value_hex, mask_hex, match_id):
        from collections import namedtuple

        Ipv6Entry = namedtuple("Ipv6Entry", "match_id octet_list mask_hex")

        OCTET_LEN = 4
        ipv6_entry_list = [
            Ipv6Entry(
                match_id=match_id,
                octet_list=[
                    value_hex[i: i + OCTET_LEN]
                    for i in range(0, len(value_hex), OCTET_LEN)
                ],
                mask_hex=mask_hex),
        ]

        while True:
            try:
                line = self.__buffer[self.__parse_idx].strip()
            except IndexError:
                break

            try:
                value_hex, mask_hex, match_id = self.__parse_filter_ip_line(
                    line)
            except pp.ParseException:
                break

            if (match_id in self.FilterMatchIdIpv6.INCOMING_NETWORK_LIST or
                    match_id in self.FilterMatchIdIpv6.OUTGOING_NETWORK_LIST):
                ipv6_entry_list.append(
                    Ipv6Entry(
                        match_id=match_id,
                        octet_list=[
                            value_hex[i: i + OCTET_LEN]
                            for i in range(0, len(value_hex), OCTET_LEN)
                        ],
                        mask_hex=mask_hex))
            else:
                break

            self.__parse_idx += 1

        src_octet_list = []
        dst_octet_list = []
        src_netmask = 0
        dst_netmask = 0

        for ipv6_entry in ipv6_entry_list:
            part_netmask = bin(int(ipv6_entry.mask_hex, 16)).count("1")

            if ipv6_entry.match_id in self.FilterMatchIdIpv6.INCOMING_NETWORK_LIST:
                src_octet_list.extend(ipv6_entry.octet_list)
                src_netmask += part_netmask
            elif ipv6_entry.match_id in self.FilterMatchIdIpv6.OUTGOING_NETWORK_LIST:
                dst_octet_list.extend(ipv6_entry.octet_list)
                dst_netmask += part_netmask
            else:
                raise ValueError(
                    "unexpected ipv6 entry: {}".format(ipv6_entry))

        while len(src_octet_list) < 8:
            src_octet_list.append("0000")
        while len(dst_octet_list) < 8:
            dst_octet_list.append("0000")

        self.__filter_dst_network = ipaddress.IPv6Network("{:s}/{:d}".format(
            ":".join(dst_octet_list), dst_netmask)).compressed
        self.__filter_src_network = ipaddress.IPv6Network("{:s}/{:d}".format(
            ":".join(src_octet_list), src_netmask)).compressed

    def __parse_filter_port(self, value_hex):
        # Port filter consists eight hex digits.
        # The upper-half represents source port filter and
        # the bottom-half represents destination port filter.

        if len(value_hex) != 8:
            raise ValueError("invalid port filter value: {}".format(value_hex))

        src_port_hex = value_hex[:4]
        dst_port_hex = value_hex[4:]

        logger.debug(
            "parse ipv4 port: src-port-hex={}, dst-port-hex={}".format(
                src_port_hex, dst_port_hex))

        src_port_decimal = int(src_port_hex, 16)
        self.__filter_src_port = (
            src_port_decimal if src_port_decimal != 0 else None)

        dst_port_decimal = int(dst_port_hex, 16)
        self.__filter_dst_port = (
            dst_port_decimal if dst_port_decimal != 0 else None)

    def __parse_filter_ipv4(self, line):
        value_hex, mask_hex, match_id = self.__parse_filter_ip_line(line)

        if match_id in [
                self.FilterMatchIdIpv4.INCOMING_NETWORK,
                self.FilterMatchIdIpv4.OUTGOING_NETWORK,
        ]:
            self.__parse_filter_ipv4_network(value_hex, mask_hex, match_id)
        elif match_id == self.FilterMatchIdIpv4.PORT:
            self.__parse_filter_port(value_hex)
        elif match_id in (self.FilterMatchIdIpv6.INCOMING_NETWORK_LIST +
                          self.FilterMatchIdIpv6.OUTGOING_NETWORK_LIST +
                          [self.FilterMatchIdIpv6.PORT]):
            logger.warn(
                "unknown match id for an IPv4 filter: "
                "might be an IPv6 filter. try to use --ipv6 option. "
                "(id={})".format(match_id))
            return
        else:
            logger.debug("unknown match id: {}".format(match_id))
            return

        logger.debug(
            "succeed to parse ipv4 filter: " + ", ".join([
                "src_network={}".format(self.__filter_src_network),
                "dst_network={}".format(self.__filter_dst_network),
                "src_port={}".format(self.__filter_src_port),
                "dst_port={}".format(self.__filter_dst_port),
                "line={}".format(line)
            ]))

    def __parse_filter_ipv6(self, line):
        value_hex, mask_hex, match_id = self.__parse_filter_ip_line(line)

        if (match_id in self.FilterMatchIdIpv6.INCOMING_NETWORK_LIST or
                match_id in self.FilterMatchIdIpv6.OUTGOING_NETWORK_LIST):
            self.__parse_filter_ipv6_network(value_hex, mask_hex, match_id)
        elif match_id == self.FilterMatchIdIpv6.PORT:
            self.__parse_filter_port(value_hex)
        else:
            logger.debug("unknown match id: {}".format(match_id))
            return

        logger.debug(
            "succeed to parse ipv6 filter: " + ", ".join([
                "src_network={}".format(self.__filter_src_network),
                "dst_network={}".format(self.__filter_dst_network),
                "src_port={}".format(self.__filter_src_port),
                "dst_port={}".format(self.__filter_dst_port),
                "line={}".format(line)
            ]))



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import
from __future__ import unicode_literals


VERSION = "0.15.0"

KILO_SIZE = 1000
LIST_MANGLE_TABLE_COMMAND = "iptables -t mangle --line-numbers -L"
IPV6_OPTION_ERROR_MSG_FORMAT = "{}. --ipv6 option required to use IPv6 address."


class Network(object):
    class Ipv4(object):
        ANYWHERE = "0.0.0.0/0"

    class Ipv6(object):
        ANYWHERE = "::/0"


class ShapingAlgorithm(object):
    HTB = "htb"
    TBF = "tbf"
    LIST = [HTB, TBF]


class Tc(object):

    class Command(object):
        TCSET = "tcset"
        TCDEL = "tcdel"
        TCSHOW = "tcshow"

    class Subcommand(object):
        CLASS = "class"
        FILTER = "filter"
        QDISC = "qdisc"
        SHOW = "show"

        LIST = (CLASS, FILTER, QDISC, SHOW)

    class Param(object):
        DEVICE = "device"
        FILTER_ID = "filter_id"
        CLASS_ID = "classid"
        DST_NETWORK = "dst-network"
        DST_PORT = "dst-port"
        FLOW_ID = "flowid"
        HANDLE = "handle"
        PARENT = "parent"
        PRIORITY = "priority"
        PROTOCOL = "protocol"
        SRC_NETWORK = "src-network"
        SRC_PORT = "src-port"

    class ValueRange(object):
        class LatencyTime(object):
            MIN = "0ms"
            MAX = "60min"

    class Min(object):
        LATENCY_TIME = "0ms"

    class Max(object):
        LATENCY_TIME = "60min"


class TcCommandOutput(object):
    NOT_SET = None
    STDOUT = "STDOUT"
    SCRIPT = "SCRIPT"


class TrafficDirection(object):
    OUTGOING = "outgoing"
    INCOMING = "incoming"
    LIST = [OUTGOING, INCOMING]



#!/usr/bin/env python
# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import
from __future__ import print_function

import errno
import sys

import logbook
import subprocrunner

from ._argparse_wrapper import ArgparseWrapper
from ._common import (
    check_tc_command_installation,
    initialize,
    is_execute_tc_command,
    normalize_tc_value,
    verify_network_interface,
    write_tc_script,
)
from ._const import (
    VERSION,
    Tc,
    TcCommandOutput,
)
from ._error import NetworkInterfaceNotFoundError
from ._logger import (
    LOG_FORMAT_STRING,
    logger,
    set_logger,
)
from .traffic_control import TrafficControl


logbook.StderrHandler(
    level=logbook.DEBUG, format_string=LOG_FORMAT_STRING).push_application()


def parse_option():
    parser = ArgparseWrapper(VERSION)

    group = parser.parser.add_argument_group("Traffic Control")
    group.add_argument(
        "-d", "--device", required=True,
        help="network device name (e.g. eth0)")
    group.add_argument(
        "-a", "--all", dest="is_delete_all", action="store_true",
        help="delete all of the shaping rules.")
    group.add_argument(
        "--id", dest="filter_id",
        help="""
        delete a shaping rule which has a specific id.
        you can get an id (filter_id) by tcshow command output.
        e.g. "filter_id": "800::801"
        """)

    parser.add_routing_group()

    return parser.parser.parse_args()


def create_tc_obj(options):
    from .parser.shaping_rule import TcShapingRuleParser
    from simplesqlite.sqlquery import SqlQuery

    if options.filter_id:
        ip_version = 6 if options.is_ipv6 else 4
        shaping_rule_parser = TcShapingRuleParser(
            device=options.device, ip_version=ip_version, logger=logger)
        shaping_rule_parser.parse()
        result = shaping_rule_parser.con.select_as_dict(
            table_name=Tc.Subcommand.FILTER,
            column_list=[
                Tc.Param.SRC_NETWORK, Tc.Param.DST_NETWORK,
                Tc.Param.SRC_PORT, Tc.Param.DST_PORT],
            where=SqlQuery.make_where(Tc.Param.FILTER_ID, options.filter_id))
        if not result:
            logger.error(
                "shaping rule not found associated with the id ({}).".format(
                    options.filter_id))
            sys.exit(1)

        filter_param = result[0]
        dst_network = filter_param.get(Tc.Param.DST_NETWORK)
        src_network = filter_param.get(Tc.Param.SRC_NETWORK)
        dst_port = filter_param.get(Tc.Param.DST_PORT)
        src_port = filter_param.get(Tc.Param.SRC_PORT)
    else:
        dst_network = options.dst_network
        src_network = options.src_network
        dst_port = options.dst_port
        src_port = options.src_port

    return TrafficControl(
        options.device, direction=options.direction,
        dst_network=dst_network, src_network=src_network,
        dst_port=dst_port, src_port=src_port,
        is_ipv6=options.is_ipv6)


def main():
    options = parse_option()

    initialize(options)

    if is_execute_tc_command(options.tc_command_output):
        check_tc_command_installation()
    else:
        subprocrunner.SubprocessRunner.default_is_dry_run = True
        set_logger(False)

    try:
        verify_network_interface(options.device)
    except NetworkInterfaceNotFoundError as e:
        logger.error(e)
        return errno.EINVAL

    subprocrunner.SubprocessRunner.clear_history()
    tc = create_tc_obj(options)
    if options.log_level == logbook.INFO:
        subprocrunner.set_log_level(logbook.ERROR)
    normalize_tc_value(tc)

    return_code = 0
    if options.is_delete_all:
        return_code = tc.delete_all_tc()
    else:
        return_code = tc.delete_tc()

    command_history = "\n".join(tc.get_command_history())

    if options.tc_command_output == TcCommandOutput.STDOUT:
        print(command_history)
        return return_code
    elif options.tc_command_output == TcCommandOutput.SCRIPT:
        set_logger(True)
        write_tc_script(
            Tc.Command.TCDEL, command_history, filename_suffix=options.device)
        return return_code

    logger.debug("command history\n{}".format(command_history))

    return return_code


if __name__ == '__main__':
    sys.exit(main())



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import print_function
from __future__ import unicode_literals

import pytest
from simplesqlite import connect_sqlite_memdb
import six

from tcconfig._const import Tc
import tcconfig.parser._filter
import tcconfig.parser._qdisc
import tcconfig.parser.shaping_rule


DEVICE = "eth0"


@pytest.fixture
def filter_parser_ipv4():
    return tcconfig.parser._filter.TcFilterParser(
        connect_sqlite_memdb(), ip_version=4)


@pytest.fixture
def filter_parser_ipv6():
    return tcconfig.parser._filter.TcFilterParser(
        connect_sqlite_memdb(), ip_version=6)


@pytest.fixture
def qdisc_parser():
    return tcconfig.parser._qdisc.TcQdiscParser(
        connect_sqlite_memdb())


class Test_TcFilterParser_parse_filter_ipv4(object):

    @pytest.mark.parametrize(["value", "expected"], [
        [None, []],
        ["", []],
        [
            six.b("""filter parent 1: protocol ip pref 1 u32
filter parent 1: protocol ip pref 1 u32 fh 801: ht divisor 1
filter parent 1: protocol ip pref 1 u32 fh 801::800 order 2048 key ht 801 bkt 0 flowid 1:1
  match c0a8000a/ffffffff at 16
filter parent 1: protocol ip pref 2 u32
filter parent 1: protocol ip pref 2 u32 fh 800: ht divisor 1
filter parent 1: protocol ip pref 2 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1:2
  match 00000000/00000000 at 16"""),
            [
                {
                    Tc.Param.DEVICE: DEVICE,
                    Tc.Param.FILTER_ID: '801::800',
                    Tc.Param.FLOW_ID: '1:1',
                    Tc.Param.SRC_NETWORK: '0.0.0.0/0',
                    Tc.Param.DST_NETWORK: '192.168.0.10/32',
                    Tc.Param.PROTOCOL: 'ip',
                    Tc.Param.PRIORITY: 1,
                    Tc.Param.SRC_PORT: None,
                    Tc.Param.DST_PORT: None,
                },
                {
                    Tc.Param.DEVICE: DEVICE,
                    Tc.Param.FILTER_ID: '800::800',
                    Tc.Param.FLOW_ID: '1:2',
                    Tc.Param.SRC_NETWORK: '0.0.0.0/0',
                    Tc.Param.DST_NETWORK: '0.0.0.0/0',
                    Tc.Param.PROTOCOL: 'ip',
                    Tc.Param.PRIORITY: 2,
                    Tc.Param.SRC_PORT: None,
                    Tc.Param.DST_PORT: None,
                },
            ],
        ],
        [
            six.b("""filter parent 1: protocol ip pref 1 u32
filter parent 1: protocol ip pref 1 u32 fh 801: ht divisor 1
filter parent 1: protocol ip pref 1 u32 fh 801::800 order 2048 key ht 801 bkt 0 flowid 1:1
  match c0a80000/ffffff00 at 16
  match 00000050/0000ffff at 20
filter parent 1: protocol ip pref 2 u32
filter parent 1: protocol ip pref 2 u32 fh 800: ht divisor 1
filter parent 1: protocol ip pref 2 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1:2
  match 00000000/00000000 at 16
  match 04d20000/ffff0000 at 20"""),
            [
                {
                    Tc.Param.DEVICE: DEVICE,
                    Tc.Param.FILTER_ID: '801::800',
                    Tc.Param.FLOW_ID: '1:1',
                    Tc.Param.SRC_NETWORK: '0.0.0.0/0',
                    Tc.Param.DST_NETWORK: '192.168.0.0/24',
                    Tc.Param.PROTOCOL: 'ip',
                    Tc.Param.PRIORITY: 1,
                    Tc.Param.SRC_PORT: None,
                    Tc.Param.DST_PORT: 80,
                },
                {
                    Tc.Param.DEVICE: DEVICE,
                    Tc.Param.FILTER_ID: '800::800',
                    Tc.Param.FLOW_ID: '1:2',
                    Tc.Param.SRC_NETWORK: '0.0.0.0/0',
                    Tc.Param.DST_NETWORK: '0.0.0.0/0',
                    Tc.Param.PROTOCOL: 'ip',
                    Tc.Param.PRIORITY: 2,
                    Tc.Param.SRC_PORT: 1234,
                    Tc.Param.DST_PORT: None,
                },
            ],
        ],
        [
            six.b("""filter parent 1: protocol ip pref 1 u32
filter parent 1: protocol ip pref 1 u32 fh 801: ht divisor 1
filter parent 1: protocol ip pref 1 u32 fh 801::800 order 2048 key ht 801 bkt 0 flowid 1:3
  match c0a8000a/ffffffff at 12
  match 00001f90/0000ffff at 20
filter parent 1: protocol ip pref 2 u32
filter parent 1: protocol ip pref 2 u32 fh 800: ht divisor 1
filter parent 1: protocol ip pref 2 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1:2
  match 00000000/00000000 at 12"""),
            [
                {
                    Tc.Param.DEVICE: DEVICE,
                    Tc.Param.FILTER_ID: '801::800',
                    Tc.Param.FLOW_ID: '1:3',
                    Tc.Param.SRC_NETWORK: '192.168.0.10/32',
                    Tc.Param.DST_NETWORK: '0.0.0.0/0',
                    Tc.Param.PROTOCOL: 'ip',
                    Tc.Param.PRIORITY: 1,
                    Tc.Param.SRC_PORT: None,
                    Tc.Param.DST_PORT: 8080,
                },
                {
                    Tc.Param.DEVICE: DEVICE,
                    Tc.Param.FILTER_ID: '800::800',
                    Tc.Param.FLOW_ID: '1:2',
                    Tc.Param.SRC_NETWORK: '0.0.0.0/0',
                    Tc.Param.DST_NETWORK: '0.0.0.0/0',
                    Tc.Param.PROTOCOL: 'ip',
                    Tc.Param.PRIORITY: 2,
                    Tc.Param.SRC_PORT: None,
                    Tc.Param.DST_PORT: None,
                },
            ],
        ],
        [
            six.b("""filter parent 1a1a: protocol ip pref 1 u32
filter parent 1a1a: protocol ip pref 1 u32 fh 800: ht divisor 1
filter parent 1a1a: protocol ip pref 1 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1a1a:2
  match 00000000/00000000 at 16
  match 15b3115c/ffffffff at 20"""),
            [
                {
                    Tc.Param.DEVICE: DEVICE,
                    Tc.Param.FILTER_ID: '800::800',
                    Tc.Param.FLOW_ID: '1a1a:2',
                    Tc.Param.SRC_NETWORK: '0.0.0.0/0',
                    Tc.Param.DST_NETWORK: '0.0.0.0/0',
                    Tc.Param.PROTOCOL: 'ip',
                    Tc.Param.PRIORITY: 1,
                    Tc.Param.SRC_PORT: 5555,
                    Tc.Param.DST_PORT: 4444,
                },
            ],
        ],
        [
            six.b("""filter parent 1f1c: protocol ip pref 1 fw
filter parent 1f1c: protocol ip pref 1 fw handle 0x65 classid 1f1c:1"""),
            [
                {
                    Tc.Param.DEVICE: DEVICE,
                    'classid': '1f1c:1',
                    'handle': 101,
                },
            ],
        ],
    ])
    def test_normal(self, filter_parser_ipv4, value, expected):
        actual = filter_parser_ipv4.parse(DEVICE, value)

        print("[expected]\n{}\n".format(expected))
        print("[actual]\n{}\n".format(actual))

        assert actual == expected


class Test_TcFilterParser_parse_filter_ipv6(object):

    @pytest.mark.parametrize(["value", "expected"], [
        [None, []],
        ["", []],
        [
            six.b("""filter parent 1f87: protocol ipv6 pref 1 u32
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800: ht divisor 1
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1f87:2
  match 00000000/ffffffff at 24
  match 00000000/ffffffff at 28
  match 00000000/ffffffff at 32
  match 00000001/ffffffff at 36"""),
            [
                {
                    Tc.Param.DEVICE: DEVICE,
                    Tc.Param.FILTER_ID: '800::800',
                    Tc.Param.FLOW_ID: '1f87:2',
                    Tc.Param.PROTOCOL: 'ipv6',
                    Tc.Param.PRIORITY: 1,
                    Tc.Param.SRC_NETWORK: '::/0',
                    Tc.Param.DST_NETWORK: '::1/128',
                    Tc.Param.SRC_PORT: None,
                    Tc.Param.DST_PORT: None,
                },
            ]
        ],
        [
            six.b("""filter parent 1f87: protocol ipv6 pref 1 u32
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800: ht divisor 1
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1f87:2
  match 2001db00/ffffffff at 8
  match 00000000/ffffffff at 12
  match 00000000/ffffffff at 16
  match 00000001/ffffffff at 20"""),
            [
                {
                    Tc.Param.DEVICE: DEVICE,
                    Tc.Param.FILTER_ID: '800::800',
                    Tc.Param.FLOW_ID: '1f87:2',
                    Tc.Param.PROTOCOL: 'ipv6',
                    Tc.Param.PRIORITY: 1,
                    Tc.Param.SRC_NETWORK: '2001:db00::1/128',
                    Tc.Param.DST_NETWORK: '::/0',
                    Tc.Param.SRC_PORT: None,
                    Tc.Param.DST_PORT: None
                }
            ]
        ],
        [
            six.b("""filter parent 1f87: protocol ipv6 pref 1 u32
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800: ht divisor 1
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1f87:2
  match 2001db00/ffffff00 at 24
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800::801 order 2049 key ht 800 bkt 0 flowid 1f87:3
  match 2001db00/ffffffff at 24
  match 00000000/ffffffff at 28
  match 00000000/ffffffff at 32
  match 00000001/ffffffff at 36
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800::802 order 2050 key ht 800 bkt 0 flowid 1f87:4
  match 00001f90/0000ffff at 40"""),
            [
                {
                    Tc.Param.DEVICE: DEVICE,
                    Tc.Param.FILTER_ID: '800::800',
                    Tc.Param.FLOW_ID: '1f87:2',
                    Tc.Param.PROTOCOL: 'ipv6',
                    Tc.Param.PRIORITY: 1,
                    Tc.Param.SRC_NETWORK: '::/0',
                    Tc.Param.DST_NETWORK: '2001:db00::/24',
                    Tc.Param.SRC_PORT: None,
                    Tc.Param.DST_PORT: None
                },
                {
                    Tc.Param.DEVICE: DEVICE,
                    Tc.Param.FILTER_ID: '800::801',
                    Tc.Param.FLOW_ID: '1f87:3',
                    Tc.Param.PROTOCOL: 'ipv6',
                    Tc.Param.PRIORITY: 1,
                    Tc.Param.SRC_NETWORK: '::/0',
                    Tc.Param.DST_NETWORK: '2001:db00::1/128',
                    Tc.Param.SRC_PORT: None,
                    Tc.Param.DST_PORT: None
                },
                {
                    Tc.Param.DEVICE: DEVICE,
                    Tc.Param.FILTER_ID: '800::802',
                    Tc.Param.FLOW_ID: '1f87:4',
                    Tc.Param.PROTOCOL: 'ipv6',
                    Tc.Param.PRIORITY: 1,
                    Tc.Param.SRC_NETWORK: '::/0',
                    Tc.Param.DST_NETWORK: '::/0',
                    Tc.Param.SRC_PORT: None,
                    Tc.Param.DST_PORT: 8080
                },
            ]
        ],
        [
            six.b("""filter parent 1f87: protocol ipv6 pref 1 u32
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800: ht divisor 1
filter parent 1f87: protocol ipv6 pref 1 u32 fh 800::802 order 2050 key ht 800 bkt 0 flowid 1f87:4
  match 00501f90/ffffffff at 40"""),
            [
                {
                    Tc.Param.DEVICE: DEVICE,
                    Tc.Param.FILTER_ID: '800::802',
                    Tc.Param.FLOW_ID: '1f87:4',
                    Tc.Param.PROTOCOL: 'ipv6',
                    Tc.Param.PRIORITY: 1,
                    Tc.Param.SRC_NETWORK: '::/0',
                    Tc.Param.DST_NETWORK: '::/0',
                    Tc.Param.SRC_PORT: 80,
                    Tc.Param.DST_PORT: 8080
                },
            ]
        ],
    ])
    def test_normal(self, filter_parser_ipv6, value, expected):
        actual = filter_parser_ipv6.parse(DEVICE, value)

        print("[expected]\n{}\n".format(expected))
        print("[actual]\n{}\n".format(actual))

        assert actual == expected


class Test_TcFilterParser_parse_incoming_device(object):

    @pytest.mark.parametrize(["value", "expected"], [
        ["", None],
        [None, None],
        [
            six.b("""filter parent ffff: protocol ip pref 49152 u32
filter parent ffff: protocol ip pref 49152 u32 fh 800: ht divisor 1
filter parent ffff: protocol ip pref 49152 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1f87:
  match 00000000/00000000 at 0
        action order 1: mirred (Egress Redirect to device ifb8071) stolen
        index 98 ref 1 bind 1"""),
            "ifb8071",
        ],
        [
            six.b("""filter parent ffff: protocol ip pref 49152 u32
filter parent ffff: protocol ip pref 49152 u32 fh 800: ht divisor 1
filter parent ffff: protocol ip pref 49152 u32 fh 800::800 order 2048 key ht 800 bkt 0 flowid 1:
  match 00000000/00000000 at 0
        action order 1: mirred (Egress Redirect to device ifb0) stolen
        index 3795 ref 1 bind 1"""),
            "ifb0",
        ],
    ])
    def test_normal(self, filter_parser_ipv4, value, expected):
        assert filter_parser_ipv4.parse_incoming_device(value) == expected


class Test_TcQdiscParser_parse(object):

    @pytest.mark.parametrize(["value", "expected"], [
        [
            six.b("""qdisc htb 1f87: root refcnt 2 r2q 10 default 1 direct_packets_stat 1 direct_qlen 1000
qdisc netem 2007: parent 1f87:2 limit 1000 delay 1.0ms loss 0.01%
"""),
            [
                {
                    Tc.Param.DEVICE: DEVICE,
                    'delay': '1.0ms', 'loss': '0.01',
                    Tc.Param.HANDLE: '2007:',
                    Tc.Param.PARENT: '1f87:2',
                },
            ],
        ],
        [
            six.b("""qdisc htb 1a1a: root refcnt 2 r2q 10 default 1 direct_packets_stat 4 direct_qlen 1000
qdisc netem 1a9a: parent 1a1a:2 limit 1000
"""),
            [
                {
                    Tc.Param.DEVICE: DEVICE,
                    Tc.Param.HANDLE: '1a9a:',
                    Tc.Param.PARENT: '1a1a:2',
                },
            ],
        ],
        [
            six.b("""
qdisc htb 1f87: root refcnt 2 r2q 10 default 1 direct_packets_stat 0 direct_qlen 1000
qdisc netem 2007: parent 1f87:2 limit 1000 delay 5.0ms
qdisc netem 2008: parent 1f87:3 limit 1000 delay 50.0ms  1.0ms loss 5%
"""),
            [
                {
                    Tc.Param.DEVICE: DEVICE,
                    'delay': '5.0ms',
                    Tc.Param.HANDLE: '2007:',
                    Tc.Param.PARENT: '1f87:2',
                },
                {
                    Tc.Param.DEVICE: DEVICE,
                    'delay': '50.0ms', 'loss': '5',
                    'delay-distro': '1.0ms',
                    Tc.Param.HANDLE: '2008:',
                    Tc.Param.PARENT: '1f87:3',
                },
            ],
        ],
        [
            six.b("""
qdisc htb 1f87: root refcnt 2 r2q 10 default 1 direct_packets_stat 0 direct_qlen 1000
qdisc netem 2007: parent 1f87:2 limit 1000 delay 2.5s
qdisc netem 2008: parent 1f87:3 limit 1000 delay 0.5s  1.0ms loss 5%
"""),
            [
                {
                    Tc.Param.DEVICE: DEVICE,
                    'delay': '2.5s',
                    Tc.Param.HANDLE: '2007:',
                    Tc.Param.PARENT: '1f87:2',
                },
                {
                    Tc.Param.DEVICE: DEVICE,
                    'delay': '0.5s', 'loss': '5',
                    'delay-distro': '1.0ms',
                    Tc.Param.HANDLE: '2008:',
                    Tc.Param.PARENT: '1f87:3',
                },
            ],
        ],
    ])
    def test_normal(self, qdisc_parser, value, expected):
        actual = qdisc_parser.parse(DEVICE, value)

        print("[expected]\n{}\n".format(expected))
        print("[actual]\n{}\n".format(actual))

        assert actual == expected



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import
from __future__ import unicode_literals

from subprocrunner import SubprocessRunner
import typepy

from .._common import (
    get_anywhere_network,
    get_no_limit_kbits,
    logging_context,
    run_command_helper,
)
from .._const import (
    ShapingAlgorithm,
    Tc,
    TrafficDirection,
)
from .._error import InvalidParameterError
from ._interface import AbstractShaper


class TbfShaper(AbstractShaper):

    __NETEM_QDISC_MAJOR_ID_OFFSET = 10

    __MIN_BUFFER_BYTE = 1600

    __OUT_DEVICE_QDISC_MINOR_ID = 1
    __IN_DEVICE_QDISC_MINOR_ID = 3

    @property
    def algorithm_name(self):
        return ShapingAlgorithm.TBF

    def _get_qdisc_minor_id(self):
        if self._tc_obj.direction == TrafficDirection.OUTGOING:
            return self.__OUT_DEVICE_QDISC_MINOR_ID

        if self._tc_obj.direction == TrafficDirection.INCOMING:
            return self.__IN_DEVICE_QDISC_MINOR_ID

        raise InvalidParameterError(
            "unknown direction",
            expected=TrafficDirection.LIST, value=self.direction)

    def _get_netem_qdisc_major_id(self, base_id):
        if self._tc_obj.direction == TrafficDirection.OUTGOING:
            direction_offset = 0
        elif self._tc_obj.direction == TrafficDirection.INCOMING:
            direction_offset = 1

        return (
            base_id +
            self.__NETEM_QDISC_MAJOR_ID_OFFSET +
            direction_offset)

    def _make_qdisc(self):
        base_command = self._tc_obj.get_tc_command(Tc.Subcommand.QDISC)
        if base_command is None:
            return 0

        handle = "{:s}:".format(self._tc_obj.qdisc_major_id_str)

        return run_command_helper(
            " ".join([
                base_command, self._dev, "root",
                "handle {:s}".format(handle), "prio",
            ]),
            self._tc_obj.REGEXP_FILE_EXISTS,
            self._tc_obj.EXISTS_MSG_TEMPLATE.format(
                "failed to '{command:s}': prio qdisc already exists "
                "({dev:s}, algo={algorithm:s}, handle={handle:s}).".format(
                    command=base_command, dev=self._dev,
                    algorithm=self.algorithm_name, handle=handle))
        )

    def _add_rate(self):
        try:
            self._tc_obj.validate_bandwidth_rate()
        except InvalidParameterError:
            return 0

        base_command = self._tc_obj.get_tc_command(Tc.Subcommand.QDISC)
        if base_command is None:
            return 0

        parent = "{:x}:{:d}".format(
            self._get_netem_qdisc_major_id(self._tc_obj.qdisc_major_id),
            self._get_qdisc_minor_id())
        handle = "{:d}:".format(20)
        no_limit_kbits = get_no_limit_kbits(self._tc_device)

        kbits = self._tc_obj.bandwidth_rate
        if kbits is None:
            kbits = no_limit_kbits

        command = " ".join([
            base_command,
            self._dev,
            "parent {:s}".format(parent),
            "handle {:s}".format(handle),
            self.algorithm_name,
            "rate {}kbit".format(kbits),
            "buffer {:d}".format(
                max(int(kbits), self.__MIN_BUFFER_BYTE)
            ),  # [byte]
            "limit 10000",
        ])

        run_command_helper(
            command,
            self._tc_obj.REGEXP_FILE_EXISTS,
            self._tc_obj.EXISTS_MSG_TEMPLATE.format(
                "failed to '{command:s}': qdisc already exists "
                "({dev:s}, algo={algorithm:s}, parent={parent:s}, "
                "handle={handle:s}).".format(
                    command=base_command, dev=self._dev,
                    algorithm=self.algorithm_name, parent=parent,
                    handle=handle))
        )

        self.__set_pre_network_filter()

    def set_shaping(self):
        with logging_context("_make_qdisc"):
            self._make_qdisc()

        with logging_context("_set_netem"):
            self._set_netem()

        with logging_context("_add_rate"):
            self._add_rate()

        with logging_context("_add_filter"):
            self._add_filter()

        return 0

    def __set_pre_network_filter(self):
        if self._is_use_iptables():
            return 0

        if all([
                typepy.is_null_string(self._tc_obj.dst_network),
                not typepy.type.Integer(self._tc_obj.dst_port).is_type(),
        ]):
            flowid = "{:s}:{:d}".format(
                self._tc_obj.qdisc_major_id_str,
                self._get_qdisc_minor_id())
        else:
            flowid = "{:s}:2".format(self._tc_obj.qdisc_major_id_str)

        return SubprocessRunner(" ".join([
            self._tc_obj.get_tc_command(Tc.Subcommand.FILTER),
            self._dev,
            "protocol {:s}".format(self._tc_obj.protocol),
            "parent {:s}:".format(self._tc_obj.qdisc_major_id_str),
            "prio 2 u32 match {:s} {:s} {:s}".format(
                self._tc_obj.protocol,
                self._get_network_direction_str(),
                get_anywhere_network(self._tc_obj.ip_version)),
            "flowid {:s}".format(flowid),
        ])).run()



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import
from __future__ import unicode_literals

import copy

import simplesqlite
import subprocrunner
import typepy
from typepy.type import Integer

from .._common import (
    is_anywhere_network,
    run_tc_show,
)
from .._const import (
    Tc,
    TrafficDirection,
)
from .._iptables import IptablesMangleController
from ._class import TcClassParser
from ._filter import TcFilterParser
from ._qdisc import TcQdiscParser


class TcShapingRuleParser(object):

    @property
    def con(self):
        return self.__con

    @property
    def device(self):
        return self.__device

    @property
    def ifb_device(self):
        return self.__ifb_device

    def __init__(self, device, ip_version, logger):
        self.__device = device
        self.__ip_version = ip_version
        self.__logger = logger

        self.clear()
        self.__ifb_device = self.__get_ifb_from_device()

        self.__iptables_ctrl = IptablesMangleController(True, ip_version)

    def clear(self):
        self.__con = simplesqlite.connect_sqlite_memdb()
        self.__filter_parser = TcFilterParser(self.__con, self.__ip_version)
        self.__parsed_mappings = {}

    def get_tc_parameter(self):
        return {
            self.device: {
                TrafficDirection.OUTGOING: self.__get_shaping_rule(
                    self.device),
                TrafficDirection.INCOMING: self.__get_shaping_rule(
                    self.ifb_device),
            },
        }

    def get_outgoing_tc_filter(self):
        return self.__parse_tc_filter(self.device)

    def get_incoming_tc_filter(self):
        if not self.ifb_device:
            return []

        return self.__parse_tc_filter(self.ifb_device)

    def parse(self):
        self.__parse_device(self.device)
        self.__parse_device(self.ifb_device)

    def __parse_device(self, device):
        if not device:
            return

        if self.__parsed_mappings.get(device):
            return

        self.__parse_tc_class(device)
        self.__parse_tc_filter(device)
        self.__parse_tc_qdisc(device)

        self.__parsed_mappings[device] = True

    def __get_ifb_from_device(self):
        filter_runner = subprocrunner.SubprocessRunner(
            "tc filter show dev {:s} root".format(self.device), dry_run=False)
        filter_runner.run()

        return self.__filter_parser.parse_incoming_device(filter_runner.stdout)

    def __get_filter_key(self, filter_param):
        src_network_format = Tc.Param.SRC_NETWORK + "={:s}"
        dst_network_format = Tc.Param.DST_NETWORK + "={:s}"
        protocol_format = Tc.Param.PROTOCOL + "={:s}"
        key_item_list = []

        if Tc.Param.HANDLE in filter_param:
            handle = filter_param.get(Tc.Param.HANDLE)
            Integer(handle).validate()
            handle = int(handle)

            for mangle in self.__iptables_ctrl.parse():
                if mangle.mark_id != handle:
                    continue

                key_item_list.append(
                    dst_network_format.format(mangle.destination))
                if typepy.is_not_null_string(mangle.source):
                    key_item_list.append("{:s}={:s}".format(
                        Tc.Param.SRC_NETWORK, mangle.source))
                key_item_list.append(protocol_format.format(mangle.protocol))

                break
            else:
                raise ValueError("mangle mark not found: {}".format(mangle))
        else:
            src_network = filter_param.get(Tc.Param.SRC_NETWORK)
            if (typepy.is_not_null_string(src_network) and
                    not is_anywhere_network(src_network, self.__ip_version)):
                key_item_list.append(src_network_format.format(src_network))

            dst_network = filter_param.get(Tc.Param.DST_NETWORK)
            if (typepy.is_not_null_string(dst_network) and
                    not is_anywhere_network(dst_network, self.__ip_version)):
                key_item_list.append(dst_network_format.format(dst_network))

            src_port = filter_param.get(Tc.Param.SRC_PORT)
            if Integer(src_port).is_type():
                port_format = Tc.Param.SRC_PORT + "={:d}"
                key_item_list.append(port_format.format(src_port))

            dst_port = filter_param.get(Tc.Param.DST_PORT)
            if Integer(dst_port).is_type():
                port_format = Tc.Param.DST_PORT + "={:d}"
                key_item_list.append(port_format.format(dst_port))

            protocol = filter_param.get(Tc.Param.PROTOCOL)
            if typepy.is_not_null_string(protocol):
                key_item_list.append(protocol_format.format(protocol))

        return ", ".join(key_item_list)

    def __get_shaping_rule(self, device):
        from simplesqlite.sqlquery import SqlQuery

        if typepy.is_null_string(device):
            return {}

        self.__parse_device(device)
        where_query = SqlQuery.make_where(Tc.Param.DEVICE, device)

        try:
            class_param_list = self.__con.select_as_dict(
                table_name=Tc.Subcommand.CLASS, where=where_query)
        except simplesqlite.TableNotFoundError:
            class_param_list = []

        try:
            filter_param_list = self.__con.select_as_dict(
                table_name=Tc.Subcommand.FILTER, where=where_query)
        except simplesqlite.TableNotFoundError:
            filter_param_list = []

        try:
            qdisc_param_list = self.__con.select_as_dict(
                table_name=Tc.Subcommand.QDISC, where=where_query)
        except simplesqlite.TableNotFoundError:
            qdisc_param_list = []

        shaping_rule_mapping = {}

        for filter_param in filter_param_list:
            self.__logger.debug(
                "{:s} param: {}".format(Tc.Subcommand.FILTER, filter_param))
            shaping_rule = {}

            filter_key = self.__get_filter_key(filter_param)
            if typepy.is_null_string(filter_key):
                self.__logger.debug(
                    "empty filter key: {}".format(filter_param))
                continue

            for qdisc_param in qdisc_param_list:
                self.__logger.debug(
                    "{:s} param: {}".format(Tc.Subcommand.QDISC, qdisc_param))

                if qdisc_param.get(Tc.Param.PARENT) not in (
                        filter_param.get(Tc.Param.FLOW_ID),
                        filter_param.get(Tc.Param.CLASS_ID)):
                    continue

                shaping_rule[Tc.Param.FILTER_ID] = filter_param.get(
                    Tc.Param.FILTER_ID)
                # shaping_rule[Tc.Param.PRIORITY] = filter_param.get(
                #    Tc.Param.PRIORITY)
                shaping_rule.update(self.__strip_param(
                    qdisc_param,
                    [Tc.Param.DEVICE, Tc.Param.PARENT, Tc.Param.HANDLE]))

            for class_param in class_param_list:
                self.__logger.debug(
                    "{:s} param: {}".format(Tc.Subcommand.CLASS, class_param))

                if class_param.get(Tc.Param.CLASS_ID) not in (
                        filter_param.get(Tc.Param.FLOW_ID),
                        filter_param.get(Tc.Param.CLASS_ID)):
                    continue

                shaping_rule[Tc.Param.FILTER_ID] = filter_param.get(
                    Tc.Param.FILTER_ID)
                # shaping_rule[Tc.Param.PRIORITY] = filter_param.get(
                #    Tc.Param.PRIORITY)
                shaping_rule.update(self.__strip_param(
                    class_param, [Tc.Param.DEVICE, Tc.Param.CLASS_ID]))

            if not shaping_rule:
                self.__logger.debug(
                    "shaping rule not found for '{}'".format(filter_param))
                continue

            self.__logger.debug(
                "shaping rule found: {} {}".format(filter_key, shaping_rule))

            shaping_rule_mapping[filter_key] = shaping_rule

        return shaping_rule_mapping

    def __parse_tc_qdisc(self, device):
        return TcQdiscParser(self.__con).parse(
            device, run_tc_show(Tc.Subcommand.QDISC, device))

    def __parse_tc_filter(self, device):
        return self.__filter_parser.parse(
            device, run_tc_show(Tc.Subcommand.FILTER, device))

    def __parse_tc_class(self, device):
        return TcClassParser(self.__con).parse(
            device, run_tc_show(Tc.Subcommand.CLASS, device))

    @staticmethod
    def __strip_param(params, strip_param_list):
        work_params = copy.deepcopy(params)

        for strip_param in strip_param_list:
            try:
                del work_params[strip_param]
            except KeyError:
                pass

        return work_params



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import json

import pytest
from subprocrunner import SubprocessRunner

from tcconfig._const import Tc

from .common import execute_tcdel


@pytest.fixture
def device_value(request):
    return request.config.getoption("--device")


class Test_tcshow(object):
    """
    Tests in this class are not executable on CI services.
    Execute the following command at the local environment to running tests:

        python setup.py test --addopts "--runxfail --device=<test device>"
    """

    def test_normal_empty(self, device_value):
        if device_value is None:
            pytest.skip("device option is null")

        execute_tcdel(device_value)

        runner = SubprocessRunner("{:s} --device {:s}".format(
            Tc.Command.TCSHOW, device_value))

        expected = "{" + '"{:s}"'.format(device_value) + ": {" + """
        "outgoing": {
        },
        "incoming": {
        }
    }
}"""
        print("[expected]\n{}\n".format(expected))
        print("[actual]\n{}\n".format(runner.stdout))

        assert runner.run() == 0
        assert json.loads(runner.stdout) == json.loads(expected)

    def test_normal_ipv4(self, device_value):
        if device_value is None:
            pytest.skip("device option is null")

        device_option = "--device {:s}".format(device_value)

        assert SubprocessRunner(" ".join([
            Tc.Command.TCSET, device_option,
            "--delay", "10",
            "--delay-distro", "2",
            "--loss", "0.01",
            "--duplicate", "0.5",
            "--reorder", "0.2",
            "--rate", "0.25K",
            "--network", "192.168.0.10",
            "--port", "8080",
            "--overwrite",
        ])).run() == 0
        assert SubprocessRunner(" ".join([
            Tc.Command.TCSET, device_option,
            "--delay", "1",
            "--loss", "1",
            "--rate", "100M",
            "--network", "192.168.1.0/24",
            "--add",
        ])).run() == 0
        assert SubprocessRunner(" ".join([
            Tc.Command.TCSET, device_option,
            "--delay", "10",
            "--delay-distro", "2",
            "--rate", "500K",
            "--direction", "incoming",
        ])).run() == 0
        assert SubprocessRunner(" ".join([
            Tc.Command.TCSET, device_option,
            "--delay", "1",
            "--loss", "0.02",
            "--duplicate", "0.5",
            "--reorder", "0.2",
            "--rate", "0.1M",
            "--network", "192.168.11.0/24",
            "--port", "80",
            "--direction", "incoming",
            "--add",
        ])).run() == 0

        runner = SubprocessRunner("{:s} {:s}".format(
            Tc.Command.TCSHOW, device_option))
        runner.run()

        expected = "{" + '"{:s}"'.format(device_value) + ": {" + """
        "outgoing": {
           "dst-network=192.168.0.10/32, dst-port=8080, protocol=ip": {
                "filter_id": "800::800",
                "delay": "10.0ms",
                "loss": 0.01,
                "duplicate": 0.5,
                "reorder": 0.2,
                "rate": 248,
                "delay-distro": "2.0ms"
            },
            "dst-network=192.168.1.0/24, protocol=ip": {
                "filter_id": "800::801",
                "delay": "1.0ms",
                "loss": 1,
                "rate": "100M"
            }
        },
        "incoming": {
            "dst-network=192.168.11.0/24, dst-port=80, protocol=ip": {
                "filter_id": "800::801",
                "delay": "1.0ms",
                "loss": 0.02,
                "duplicate": 0.5,
                "reorder": 0.2,
                "rate": "100K"
            },
            "protocol=ip": {
                "filter_id": "800::800",
                "delay": "10.0ms",
                "delay-distro": "2.0ms",
                "rate": "500K"
            }
        }
    }
}"""

        print("[expected]\n{}\n".format(expected))
        print("[actual]\n{}\n".format(runner.stdout))

        assert json.loads(runner.stdout) == json.loads(expected)

        execute_tcdel(device_value)

    def test_normal_ipv6(self, device_value):
        if device_value is None:
            pytest.skip("device option is null")

        device_option = "--device {:s}".format(device_value)

        assert SubprocessRunner(" ".join([
            Tc.Command.TCSET, device_option,
            "--delay", "10",
            "--delay-distro", "2",
            "--loss", "0.01",
            "--duplicate", "5",
            "--reorder", "2",
            "--rate", "0.25K",
            "--network", "::1",
            "--port", "8080",
            "--overwrite",
            "--ipv6",
        ])).run() == 0
        assert SubprocessRunner(" ".join([
            Tc.Command.TCSET, device_option,
            "--delay", "1",
            "--loss", "1",
            "--rate", "100M",
            "--network", "2001:db00::0/24",
            "--add",
            "--ipv6",
        ])).run() == 0
        assert SubprocessRunner(" ".join([
            Tc.Command.TCSET, device_option,
            "--delay", "10",
            "--delay-distro", "2",
            "--rate", "500K",
            "--direction", "incoming",
            "--ipv6",
        ])).run() == 0
        assert SubprocessRunner(" ".join([
            Tc.Command.TCSET, device_option,
            "--delay", "1",
            "--loss", "0.02",
            "--duplicate", "5",
            "--reorder", "2",
            "--rate", "0.1M",
            "--network", "2001:db00::0/25",
            "--port", "80",
            "--direction", "incoming",
            "--add",
            "--ipv6",
        ])).run() == 0

        runner = SubprocessRunner("tcshow {:s} --ipv6".format(device_option))
        runner.run()

        expected = "{" + '"{:s}"'.format(device_value) + ": {" + """
        "outgoing": {
            "dst-network=::1/128, dst-port=8080, protocol=ipv6": {
                "filter_id": "800::800",
                "delay": "10.0ms",
                "loss": 0.01,
                "duplicate": 5,
                "reorder": 2,
                "rate": 248,
                "delay-distro": "2.0ms"
            },
            "dst-network=2001:db00::/24, protocol=ipv6": {
                "filter_id": "800::801",
                "delay": "1.0ms",
                "loss": 1,
                "rate": "100M"
            }
        },
        "incoming": {
            "dst-network=2001:db00::/25, dst-port=80, protocol=ipv6": {
                "filter_id": "800::801",
                "delay": "1.0ms",
                "loss": 0.02,
                "duplicate": 5,
                "reorder": 2,
                "rate": "100K"
            },
            "protocol=ipv6": {
                "filter_id": "800::800",
                "delay": "10.0ms",
                "rate": "500K",
                "delay-distro": "2.0ms"
            }
        }
    }
}"""

        print("[expected]\n{}\n".format(expected))
        print("[actual]\n{}\n".format(runner.stdout))

        assert json.loads(runner.stdout) == json.loads(expected)

        execute_tcdel(device_value)



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

import pytest

from tcconfig._common import (
    is_anywhere_network,
    get_anywhere_network,
    get_iproute2_upper_limite_rate,
    get_no_limit_kbits,
    sanitize_network,
)


@pytest.fixture
def device_option(request):
    return request.config.getoption("--device")


class Test_is_anywhere_network(object):

    @pytest.mark.parametrize(["network", "ip_version", "expected"], [
        ["0.0.0.0/0", 4, True],
        ["192.168.0.0/0", 4, False],
        ["::/0", 6, True],
        ["0:0:0:0:0:0:0:0/0", 6, True],
        ["2001:db00::0/24", 6, False],
    ])
    def test_normal(self, network, ip_version, expected):
        assert is_anywhere_network(network, ip_version) == expected

    @pytest.mark.parametrize(["network", "ip_version",  "expected"], [
        [None, 4, ValueError],
        ["0.0.0.0/0", 5, ValueError],
        ["0.0.0.0/0", None, ValueError],
    ])
    def test_exception(self, network, ip_version, expected):
        with pytest.raises(expected):
            is_anywhere_network(network, ip_version)


class Test_get_iproute2_upper_limite_rate(object):

    def test_normal(self):
        assert get_iproute2_upper_limite_rate() == 32000000


class Test_get_anywhere_network(object):

    @pytest.mark.parametrize(["value", "expected"], [
        [4, "0.0.0.0/0"],
        ["4", "0.0.0.0/0"],
        [6, "::/0"],
        ["6", "::/0"],
    ])
    def test_normal(self, value, expected):
        assert get_anywhere_network(value) == expected

    @pytest.mark.parametrize(["value", "expected"], [
        [None, ValueError],
        ["test", ValueError],
    ])
    def test_exception(self, value, expected):
        with pytest.raises(expected):
            get_anywhere_network(value)


class Test_get_no_limit_kbits(object):

    @pytest.mark.parametrize(["speed", "expected"], [
        [1, 1000],
        [0, 0],
    ])
    def test_normal(self, monkeypatch, device_option, speed, expected):
        if device_option is None:
            pytest.skip("device option is null")

        monkeypatch.setattr(
            "tcconfig._common.read_iface_speed", lambda x: speed)

        assert get_no_limit_kbits(device_option) == expected

    @pytest.mark.parametrize(["speed", "expected"], [
        [-1, get_iproute2_upper_limite_rate()],
    ])
    def test_normal_paravirt(
            self, monkeypatch, device_option, speed, expected):
        if device_option is None:
            pytest.skip("device option is null")

        monkeypatch.setattr(
            "tcconfig._common.read_iface_speed", lambda x: speed)

        assert get_no_limit_kbits(device_option) == expected

    @staticmethod
    def raise_ioerror(tc_device):
        raise IOError()

    def test_exception(
            self, monkeypatch, device_option):
        if device_option is None:
            pytest.skip("device option is null")

        monkeypatch.setattr(
            "tcconfig._common.read_iface_speed", self.raise_ioerror)

        assert get_no_limit_kbits(
            device_option) == get_iproute2_upper_limite_rate()


class Test_sanitize_network(object):

    @pytest.mark.parametrize(["value", "ip_version", "expected"], [
        ["192.168.0.1", 4, "192.168.0.1/32"],
        ["192.168.0.1/32", 4, "192.168.0.1/32"],
        ["192.168.0.0/24", 4, "192.168.0.0/24"],
        ["192.168.0.0/23", 4, "192.168.0.0/23"],
        ["2001:db00::0/24", 6, "2001:db00::/24"],
        ["anywhere", 4, "0.0.0.0/0"],
        ["ANYWHERE", 4, "0.0.0.0/0"],
        ["anywhere", 6, "::/0"],
        ["ANYWHERE", 6, "::/0"],
        ["", 4, "0.0.0.0/0"],
        [None, 4, "0.0.0.0/0"],
        ["", 6, "::/0"],
        [None, 6, "::/0"],
    ])
    def test_normal(self, value, ip_version, expected):
        assert sanitize_network(value, ip_version) == expected

    @pytest.mark.parametrize(["value", "ip_version", "expected"], [
        ["192.168.0.1111", 4, ValueError],
        ["192.168.0.", 4, ValueError],
        [".168.0.1", 4, ValueError],
        ["192.168.1.0/23", 4, ValueError],
        ["192.168.0.0/33", 4, ValueError],
        ["192.168.1.0/24", 6, ValueError],
        ["2001:db00::0/24", 4, ValueError],
    ])
    def test_exception(self, value, ip_version, expected):
        with pytest.raises(expected):
            sanitize_network(value, ip_version)



#!/usr/bin/env python
# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import
from __future__ import print_function

from simplesqlite.sqlquery import SqlQuery

from ._common import is_anywhere_network
from ._const import (
    Tc,
    TrafficDirection,
)
from .parser.shaping_rule import TcShapingRuleParser


class TcShapingRuleFinder(object):

    @property
    def _parser(self):
        self.__shaping_rule_parser.parse()

        return self.__shaping_rule_parser

    def __init__(self, logger, tc):
        self.__logger = logger
        self.__tc = tc
        self.__shaping_rule_parser = TcShapingRuleParser(
            device=self.__tc.device, ip_version=self.__tc.ip_version,
            logger=self.__logger)

    def clear(self):
        self.__shaping_rule_parser.clear()

    def find_qdisc_handle(self, parent):
        return self._parser.con.get_value(
            select=Tc.Param.HANDLE,
            table_name=Tc.Subcommand.QDISC,
            where=SqlQuery.make_where(Tc.Param.PARENT, parent))

    def find_filter_param(self):
        import simplesqlite

        where_list = self.__get_filter_where_condition_list()
        table_name = Tc.Subcommand.FILTER

        try:
            result = self._parser.con.select_as_dict(
                column_list=[
                    Tc.Param.FILTER_ID, Tc.Param.PRIORITY, Tc.Param.PROTOCOL],
                table_name=table_name,
                where=" AND ".join(where_list))
        except simplesqlite.TableNotFoundError:
            return None

        if not result:
            return None

        param = result[0]
        self.__logger.debug(
            "find filter param: result={}, table={}, where={}".format(
                param, table_name, where_list))

        return param

    def find_parent(self):
        where_list = self.__get_filter_where_condition_list()
        table_name = Tc.Subcommand.FILTER
        parent = self._parser.con.get_value(
            select=Tc.Param.FLOW_ID,
            table_name=table_name,
            where=" AND ".join(where_list))

        self.__logger.debug(
            "find parent: result={}, table={}, where={}".format(
                parent, table_name, where_list))

        return parent

    def is_exist_rule(self):
        return self.find_parent() is not None

    def is_any_filter(self):
        num_records = self._parser.con.get_num_records(
            table_name=Tc.Subcommand.FILTER)

        return num_records and num_records > 0

    def is_empty_filter_condition(self):
        from typepy import is_empty_string

        return all([
            is_anywhere_network(self.__tc.dst_network, self.__tc.ip_version),
            is_anywhere_network(self.__tc.src_network, self.__tc.ip_version),
            is_empty_string(self.__tc.dst_port),
            is_empty_string(self.__tc.src_port),
        ])

    def get_parsed_device(self):
        if self.__tc.direction == TrafficDirection.OUTGOING:
            device = self._parser.device
        elif self.__tc.direction == TrafficDirection.INCOMING:
            device = self._parser.ifb_device

        return device

    def __get_filter_where_condition_list(self):
        if self.__tc.direction == TrafficDirection.OUTGOING:
            device = self._parser.device
        elif self.__tc.direction == TrafficDirection.INCOMING:
            device = self._parser.ifb_device

        return [
            SqlQuery.make_where(Tc.Param.DEVICE, device),
            SqlQuery.make_where(Tc.Param.PROTOCOL, self.__tc.protocol),
            SqlQuery.make_where(Tc.Param.DST_NETWORK, self.__tc.dst_network),
            SqlQuery.make_where(Tc.Param.SRC_NETWORK, self.__tc.src_network),
            SqlQuery.make_where(Tc.Param.DST_PORT, self.__tc.dst_port),
            SqlQuery.make_where(Tc.Param.SRC_PORT, self.__tc.src_port),
        ]



# encoding: utf-8


def pytest_addoption(parser):
    parser.addoption("--device", default=None)
    parser.addoption("--local-host", default="")
    parser.addoption("--dst-host", default="")
    parser.addoption("--dst-host-ex", default="")



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import

import json

import pytest
from subprocrunner import SubprocessRunner

from tcconfig._const import Tc

from .common import execute_tcdel


@pytest.fixture
def device_value(request):
    return request.config.getoption("--device")


class Test_tcset_change(object):
    """
    Tests in this class are not executable on CI services.
    Execute the following command at the local environment to running tests:

        python setup.py test --addopts "--device=<test device>"

    These tests expected to execute in the following environment:
       - Linux w/ iputils-ping package
       - English locale (for parsing ping output)
    """

    def test_normal(self, device_value):
        if device_value is None:
            pytest.skip("device is null")

        device_option = "--device {:s}".format(device_value)

        command_list = [
            Tc.Command.TCSET, device_option,
            "--delay 100ms --rate 50k --network 192.168.1.2 --overwrite",
        ]
        assert SubprocessRunner(" ".join(command_list)).run() == 0

        command_list = [
            Tc.Command.TCSET, device_option,
            "--delay 200.0ms",
            "--delay-distro 20",
            "--rate 100k",
            "--loss 0.01",
            "--duplicate 5",
            "--reorder 2",
            "--network 192.168.1.3",
            "--add",
        ]
        assert SubprocessRunner(" ".join(command_list)).run() == 0

        runner = SubprocessRunner("{:s} {:s}".format(
            Tc.Command.TCSHOW, device_option))
        runner.run()

        expected = "{" + '"{:s}"'.format(device_value) + ": {" + """
        "outgoing": {
            "dst-network=192.168.1.2/32, protocol=ip": {
                "filter_id": "800::800",
                "delay": "100.0ms",
                "rate": "50K"
            },
            "dst-network=192.168.1.3/32, protocol=ip": {
                "filter_id": "800::801",
                "delay": "200.0ms",
                "loss": 0.01,
                "duplicate": 5,
                "delay-distro": "20.0ms",
                "rate": "100K",
                "reorder": 2
            }
        },
        "incoming": {}
    }
}"""

        print("[expected]\n{}\n".format(expected))
        print("[actual]\n{}\n".format(runner.stdout))
        assert json.loads(runner.stdout) == json.loads(expected)

        command_list = [
            Tc.Command.TCSET, device_option,
            "--delay 300ms",
            "--delay-distro 30",
            "--rate 200k",
            "--loss 0.02",
            "--duplicate 5.5",
            "--reorder 0.2",
            "--network 192.168.1.3",
            "--change",
        ]
        assert SubprocessRunner(" ".join(command_list)).run() == 0

        runner = SubprocessRunner("{:s} {:s}".format(
            Tc.Command.TCSHOW, device_option))
        runner.run()

        expected = "{" + '"{:s}"'.format(device_value) + ": {" + """
        "outgoing": {
            "dst-network=192.168.1.2/32, protocol=ip": {
                "filter_id": "800::800",
                "delay": "100.0ms",
                "rate": "50K"
            },
            "dst-network=192.168.1.3/32, protocol=ip": {
                "filter_id": "800::801",
                "delay": "300.0ms",
                "loss": 0.02,
                "duplicate": 5.5,
                "delay-distro": "30.0ms",
                "rate": "200K",
                "reorder": 0.2
            }
        },
        "incoming": {}
    }
}"""

        print("[expected]\n{}\n".format(expected))
        print("[actual]\n{}\n".format(runner.stdout))
        assert json.loads(runner.stdout) == json.loads(expected)

        # finalize ---
        execute_tcdel(device_option)



#!/usr/bin/env python
# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import
import argparse

import logbook

from ._const import (
    TcCommandOutput,
    TrafficDirection,
)


class ArgparseWrapper(object):
    """
    Wrapper class for argparse
    """

    def __init__(self, version, description=""):
        epilog = "Issue tracker: https://github.com/thombashi/tcconfig/issues"

        self.parser = argparse.ArgumentParser(
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=description, epilog=epilog)
        self.parser.add_argument(
            "--version", action="version", version="%(prog)s " + version)

        self._add_tc_command_arg_group()
        self._add_log_level_argument_group()

        group = self.parser.add_argument_group("Debug")
        group.add_argument(
            "--stacktrace", dest="is_output_stacktrace",
            action="store_true", default=False,
            help="""print stack trace for debug information.
            --debug option required to see the debug print.
            """)

    def add_routing_group(self):
        group = self.parser.add_argument_group("Routing")
        group.add_argument(
            "--direction", choices=TrafficDirection.LIST,
            default=TrafficDirection.OUTGOING,
            help="""the direction of network communication that impose traffic control.
            'incoming' requires Linux kernel version 2.6.20 or later.
            (default = %(default)s)
            """)
        group.add_argument(
            "--network", "--dst-network", dest="dst_network",
            help="target IP-address/network to control traffic")
        group.add_argument(
            "--src-network",
            help="""set a traffic shaping rule to specific packets that routed from
            --src-network to --dst-network. this option required to execute
            with the --iptables option when you use tbf.
            the shaping rule only affect to outgoing packets
            (no effect to if you execute with "--direction incoming" option)
            """)
        group.add_argument(
            "--port", "--dst-port", dest="dst_port", type=int,
            help="target destination port number to control traffic.")
        group.add_argument(
            "--src-port", type=int,
            help="target source port number to control traffic.")
        group.add_argument(
            "--ipv6", dest="is_ipv6", action="store_true", default=False,
            help="apply traffic control to IPv6 packets rather than IPv4.")

        return group

    def _add_log_level_argument_group(self):
        dest = "log_level"

        group = self.parser.add_mutually_exclusive_group()
        group.add_argument(
            "--debug", dest=dest, action="store_const",
            const=logbook.DEBUG, default=logbook.INFO,
            help="for debug print.")
        group.add_argument(
            "--quiet", dest=dest, action="store_const",
            const=logbook.NOTSET, default=logbook.INFO,
            help="suppress execution log messages.")

        return group

    def _add_tc_command_arg_group(self):
        group = self.parser.add_mutually_exclusive_group()
        group.add_argument(
            "--tc-command", dest="tc_command_output", action="store_const",
            const=TcCommandOutput.STDOUT, default=TcCommandOutput.NOT_SET,
            help="""
            display tc commands to be executed and exit.
            these commands are not executed.
            """)

        group.add_argument(
            "--tc-script", dest="tc_command_output", action="store_const",
            const=TcCommandOutput.SCRIPT, default=TcCommandOutput.NOT_SET,
            help="""
            generate a script file that described tc commands which equivalent
            with execution tcconfig command. the script can be execute without
            tcconfig package installation.
            """)



# encoding: utf-8

"""
.. codeauthor:: Tsuyoshi Hombashi <tsuyoshi.hombashi@gmail.com>
"""

from __future__ import absolute_import

import pytest
from subprocrunner import SubprocessRunner

from tcconfig._const import Tc

from .common import execute_tcdel


@pytest.fixture
def device_option(request):
    return request.config.getoption("--device")


class Test_tcset_iface_speed(object):

    @pytest.mark.parametrize(["speed"], [
        [1],
        [-1],  # para-virtualized network driver
    ])
    def test_smoke_speed(self, monkeypatch, device_option, speed):
        if device_option is None:
            pytest.skip("device option is null")

        execute_tcdel(device_option)
        monkeypatch.setattr(
            "tcconfig._common.read_iface_speed", lambda x: speed)

        command_list = [
            Tc.Command.TCSET,
            "--device {:s}".format(device_option),
            "--rate {:s}".format("1kbps"),
        ]
        assert SubprocessRunner(" ".join(command_list)).run() == 0

        # finalize ---
        execute_tcdel(device_option)

    @pytest.mark.parametrize(["rate"], [
        ["0kbps"],
        ["99Gbps"],
    ])
    def test_abnormal(self, monkeypatch, device_option, rate):
        if device_option is None:
            pytest.skip("device option is null")

        execute_tcdel(device_option)
        monkeypatch.setattr(
            "tcconfig._common.read_iface_speed", lambda x: "1")

        command_list = [
            Tc.Command.TCSET,
            "--device {:s}".format(device_option),
            "--rate {:s}".format(rate),
        ]
        assert SubprocessRunner(" ".join(command_list)).run() != 0

        # finalize ---
        execute_tcdel(device_option)
